"""Test methods for cg/cli/set/sample"""
import pytest

from cg.cli.set.base import sample
from cg.store import Store

SUCCESS = 0


def test_invalid_sample(cli_runner, base_context):
    # GIVEN an empty database

    # WHEN running set with a sample that does not exist
    sample_id = "dummy_sample_id"
    result = cli_runner.invoke(sample, [sample_id], obj=base_context)

    # THEN then it should complain on invalid sample
    assert result.exit_code != SUCCESS


def test_skip_lims(cli_runner, base_context, base_store: Store, helpers):
    # GIVEN a database with a sample

    sample_obj = helpers.add_sample(base_store, gender="female")
    key = "name"
    new_value = "new_value"

    # WHEN setting sample but skipping lims
    result = cli_runner.invoke(
        sample,
        [sample_obj.internal_id, "-kv", "name", "dummy_value", "-y", "--skip-lims"],
        obj=base_context,
    )

    # THEN update sample should have no recorded update key and value
    assert result.exit_code == SUCCESS
    assert base_context["lims"].get_updated_sample_key() != key
    assert base_context["lims"].get_updated_sample_value() != new_value


def test_help_without_sample(cli_runner, base_context, base_store: Store, helpers):
    # GIVEN a database with no sample

    # WHEN setting sample but asking for help
    result = cli_runner.invoke(sample, ["--help"], obj=base_context)

    # THEN it should fail on not having a sample as argument
    assert result.exit_code != SUCCESS

    # THEN the flags should have been mentioned in the output
    assert "-kv" in result.output
    assert "--skip-lims" in result.output
    assert "-y" in result.output

    # THEN the name property should have been mentioned
    assert "name" in result.output


def test_help_with_sample(cli_runner, base_context, base_store: Store, helpers):
    # GIVEN a database with a sample

    sample_obj = helpers.add_sample(base_store, gender="female")

    # WHEN setting sample but skipping lims
    result = cli_runner.invoke(sample, [sample_obj.internal_id, "--help"], obj=base_context)

    # THEN it should not fail on not having a sample as argument
    assert result.exit_code == SUCCESS

    # THEN the flags should have been mentioned in the output
    assert "-kv" in result.output
    assert "--skip-lims" in result.output
    assert "-y" in result.output

    # THEN the name property should have been mentioned
    assert "name" in result.output

    # THEN the name value should have been mentioned
    assert sample_obj.name in result.output


@pytest.mark.parametrize("key", ["name", "capture_kit"])
def test_set_sample(cli_runner, base_context, base_store: Store, key, helpers):
    # GIVEN a database with a sample

    sample_obj = helpers.add_sample(base_store, gender="female")
    new_value = "new_value"
    assert getattr(sample_obj, key) != new_value

    # WHEN setting key on sample to new_value
    result = cli_runner.invoke(
        sample, [sample_obj.internal_id, "-kv", key, new_value, "-y"], obj=base_context
    )

    # THEN then it should have new_value as attribute key on the sample and in LIMS
    assert result.exit_code == SUCCESS
    assert getattr(sample_obj, key) == new_value
    assert base_context["lims"].get_updated_sample_key() == key
    assert base_context["lims"].get_updated_sample_value() == new_value


def test_sex(cli_runner, base_context, base_store: Store, helpers):
    # GIVEN a database with a sample

    sample_obj = helpers.add_sample(base_store, gender="female")
    key = "sex"
    new_value = "male"
    assert getattr(sample_obj, key) != new_value

    # WHEN setting key on sample to new_value
    result = cli_runner.invoke(
        sample, [sample_obj.internal_id, "-kv", key, new_value, "-y"], obj=base_context
    )

    # THEN then it should have new_value as attribute key on the sample and in LIMS
    assert result.exit_code == SUCCESS
    assert getattr(sample_obj, key) == new_value
    assert base_context["lims"].get_updated_sample_key() == key
    assert base_context["lims"].get_updated_sample_value() == new_value


def test_invalid_customer(cli_runner, base_context, base_store: Store, helpers):
    # GIVEN a database with a sample
    sample_id = helpers.add_sample(base_store).internal_id
    customer_id = "dummy_customer_id"
    assert base_store.Sample.query.first().customer.internal_id != customer_id

    # WHEN calling set sample with an invalid customer
    result = cli_runner.invoke(
        sample, [sample_id, "-kv", "customer", customer_id, "-y", "--skip-lims"], obj=base_context
    )

    # THEN then it should error about missing customer instead of setting the value
    assert result.exit_code != SUCCESS
    assert base_store.Sample.query.first().customer.internal_id != customer_id


def test_customer(cli_runner, base_context, base_store: Store, helpers):
    # GIVEN a database with a sample and two customers
    sample_id = helpers.add_sample(base_store).internal_id
    customer_id = helpers.ensure_customer(base_store, "another_customer").internal_id
    assert base_store.Sample.query.first().customer.internal_id != customer_id

    # WHEN calling set sample with a valid customer
    result = cli_runner.invoke(
        sample, [sample_id, "-kv", "customer", customer_id, "-y", "--skip-lims"], obj=base_context
    )

    # THEN then it should set the customer of the sample
    assert result.exit_code == SUCCESS
    assert base_store.Sample.query.first().customer.internal_id == customer_id


def test_invalid_downsampled_to(cli_runner, base_context):
    # GIVEN a database with a sample
    downsampled_to = "downsampled_to"

    # WHEN calling set sample with an invalid value of downsampled to
    result = cli_runner.invoke(
        sample, ["dummy_sample_id", "-kv", "downsampled_to", downsampled_to, "-y"], obj=base_context
    )

    # THEN wrong data type
    assert result.exit_code != SUCCESS


def test_downsampled_to(cli_runner, base_context, base_store: Store, helpers):
    # GIVEN a database with a sample
    sample_id = helpers.add_sample(base_store).internal_id
    downsampled_to = 111111
    assert base_store.Sample.query.first().downsampled_to != downsampled_to

    # WHEN calling set sample with a valid value of downsampled to
    result = cli_runner.invoke(
        sample, [sample_id, "-kv", "downsampled_to", downsampled_to, "-y"], obj=base_context
    )

    # THEN then the value should have been set on the sample
    assert result.exit_code == SUCCESS
    assert base_store.Sample.query.first().downsampled_to == downsampled_to


def test_reset_downsampled_to(cli_runner, base_context, base_store: Store, helpers):
    # GIVEN a database with a sample
    sample_id = helpers.add_sample(base_store).internal_id
    downsampled_to = 0
    assert base_store.Sample.query.first().downsampled_to != downsampled_to

    # WHEN calling set sample with a valid reset value of downsampled to
    result = cli_runner.invoke(
        sample, [sample_id, "-kv", "downsampled_to", "", "-y"], obj=base_context
    )

    # THEN then the value should have been set on the sample
    assert result.exit_code == SUCCESS
    assert not base_store.Sample.query.first().downsampled_to


def test_invalid_application(cli_runner, base_context, base_store: Store, helpers):
    # GIVEN a database with a sample
    sample_id = helpers.add_sample(base_store).internal_id
    application_tag = "dummy_application"
    assert base_store.Sample.query.first().application_version.application.tag != application_tag

    # WHEN calling set sample with an invalid application
    result = cli_runner.invoke(
        sample,
        [sample_id, "-kv", "application_version", application_tag, "-y", "--skip-lims"],
        obj=base_context,
    )

    # THEN then it should error about missing application instead of setting the value
    assert result.exit_code != SUCCESS
    assert base_store.Sample.query.first().application_version.application.tag != application_tag


def test_application(cli_runner, base_context, base_store: Store, helpers):
    # GIVEN a database with a sample and two applications
    sample_obj = helpers.add_sample(base_store)
    application_tag = helpers.ensure_application_version(
        base_store, "another_application"
    ).application.tag
    assert base_store.Sample.query.first().application_version.application.tag != application_tag

    # WHEN calling set sample with an invalid application
    result = cli_runner.invoke(
        sample,
        [
            sample_obj.internal_id,
            "-kv",
            "application_version",
            application_tag,
            "-y",
            "--skip-lims",
        ],
        obj=base_context,
    )

    # THEN then the application should have been set
    assert result.exit_code == SUCCESS
    assert sample_obj.application_version.application.tag == application_tag



"""Code for store part of CLI"""

import logging

import click

from cg.apps import crunchy, hk
from cg.cli.compress.helpers import get_fastq_individuals, update_compress_api
from cg.exc import CaseNotFoundError
from cg.meta.compress import CompressAPI
from cg.store import Store

LOG = logging.getLogger(__name__)


@click.group()
@click.pass_context
def store(context):
    """Command for storing files"""
    hk_api = hk.HousekeeperAPI(context.obj)
    crunchy_api = crunchy.CrunchyAPI(context.obj)

    compress_api = CompressAPI(hk_api=hk_api, crunchy_api=crunchy_api)
    context.obj["compress"] = compress_api
    context.obj["db"] = Store(context.obj.get("database"))


@store.command("fastq")
@click.pass_context
@click.option("-d", "--dry-run", is_flag=True)
@click.argument("case-id")
def fastq_cmd(context, case_id, dry_run):
    """Store fastq files for a case in Housekeeper"""
    LOG.info("Running store fastq")

    compress_api = context.obj["compress"]
    cg_store = context.obj["db"]

    update_compress_api(compress_api, dry_run=dry_run)

    samples = get_fastq_individuals(cg_store, case_id)

    stored_individuals = 0
    try:
        for sample_id in samples:
            was_added = compress_api.add_decompressed_fastq(sample_id)
            if was_added is False:
                LOG.info("skipping individual %s", sample_id)
                continue
            stored_individuals += 1
    except CaseNotFoundError:
        return

    LOG.info("Stored fastq files for %s individuals", stored_individuals)



"""This script tests the run cli command"""
import logging
from pathlib import Path

from cg.cli.workflow.balsamic.base import run

EXIT_SUCCESS = 0


def test_without_options(cli_runner, balsamic_context: dict):
    """Test command without case_id argument"""
    # GIVEN no case_id
    # WHEN dry running without anything specified
    result = cli_runner.invoke(run, obj=balsamic_context)
    # THEN command should NOT execute successfully
    assert result.exit_code != EXIT_SUCCESS
    # THEN command should mention argument
    assert "Missing argument" in result.output


def test_with_missing_case(cli_runner, balsamic_context: dict, caplog):
    """Test command with invalid case to start with"""
    caplog.set_level(logging.ERROR)
    # GIVEN case_id not in database
    case_id = "soberelephant"
    assert not balsamic_context["BalsamicAnalysisAPI"].store.family(case_id)
    # WHEN running
    result = cli_runner.invoke(run, [case_id], obj=balsamic_context)
    # THEN command should NOT successfully call the command it creates
    assert result.exit_code != EXIT_SUCCESS
    # THEN ERROR log should be printed containing invalid case_id
    assert case_id in caplog.text
    assert "not found" in caplog.text


def test_without_samples(cli_runner, balsamic_context: dict, caplog):
    """Test command with case_id and no samples"""
    caplog.set_level(logging.ERROR)
    # GIVEN case-id
    case_id = "no_sample_case"
    # WHEN dry running with dry specified
    result = cli_runner.invoke(run, [case_id, "--dry-run"], obj=balsamic_context)
    # THEN command should NOT execute successfully
    assert result.exit_code != EXIT_SUCCESS
    # THEN warning should be printed that no config file is found
    assert case_id in caplog.text
    assert "0" in caplog.text


def test_without_config(cli_runner, balsamic_context: dict, caplog):
    """Test command with case_id and no config file"""
    caplog.set_level(logging.ERROR)
    # GIVEN case-id
    case_id = "balsamic_case_wgs_single"
    # WHEN dry running with dry specified
    result = cli_runner.invoke(run, [case_id, "--dry-run"], obj=balsamic_context)
    # THEN command should NOT execute successfully
    assert result.exit_code != EXIT_SUCCESS
    # THEN warning should be printed that no config file is found
    assert "No config file found" in caplog.text


def test_with_config(tmpdir_factory, cli_runner, balsamic_context: dict, caplog):
    """Test command with case_id and config file"""
    caplog.set_level(logging.INFO)
    # GIVEN case-id
    case_id = "balsamic_case_wgs_single"
    # WHEN ensuring case config exists where it should be stored
    Path.mkdir(
        Path(balsamic_context["BalsamicAnalysisAPI"].get_config_path(case_id)).parent, exist_ok=True
    )
    Path(balsamic_context["BalsamicAnalysisAPI"].get_config_path(case_id)).touch(exist_ok=True)
    # WHEN dry running with dry specified
    result = cli_runner.invoke(run, [case_id, "--dry-run"], obj=balsamic_context)
    # THEN command should NOT execute successfully
    assert result.exit_code == EXIT_SUCCESS
    # THEN warning should be printed that no config file is found
    assert "balsamic" in caplog.text


def test_run_analysis(cli_runner, balsamic_context: dict, caplog):
    """Test command with run-analysis option"""
    caplog.set_level(logging.INFO)
    # GIVEN case-id
    case_id = "balsamic_case_wgs_single"
    # WHEN ensuring case config exists where it should be stored
    Path.mkdir(
        Path(balsamic_context["BalsamicAnalysisAPI"].get_config_path(case_id)).parent, exist_ok=True
    )
    Path(balsamic_context["BalsamicAnalysisAPI"].get_config_path(case_id)).touch(exist_ok=True)
    # WHEN dry running with run analysis option specified
    result = cli_runner.invoke(run, [case_id, "--dry-run", "--run-analysis"], obj=balsamic_context)
    # THEN command should execute successfully
    assert result.exit_code == EXIT_SUCCESS
    # THEN dry-print should include the option
    assert "--run-analysis" in caplog.text


def test_analysis_type_qc(cli_runner, balsamic_context: dict, caplog):
    """Test command with analysis-type qc option"""
    caplog.set_level(logging.INFO)
    # GIVEN case-id
    case_id = "balsamic_case_wgs_single"
    # WHEN ensuring case config exists where it should be stored
    Path.mkdir(
        Path(balsamic_context["BalsamicAnalysisAPI"].get_config_path(case_id)).parent, exist_ok=True
    )
    Path(balsamic_context["BalsamicAnalysisAPI"].get_config_path(case_id)).touch(exist_ok=True)
    # WHEN dry running with analysis type qc option specified
    result = cli_runner.invoke(
        run, [case_id, "--dry-run", "--analysis-type", "qc"], obj=balsamic_context
    )
    # THEN command should execute successfully
    assert result.exit_code == EXIT_SUCCESS
    # THEN dry-print should include the option
    assert "--analysis-type" in caplog.text
    assert "qc" in caplog.text


def test_priority_custom(cli_runner, balsamic_context: dict, caplog):
    """Test command with priority option"""
    caplog.set_level(logging.INFO)
    # GIVEN valid case-id
    case_id = "balsamic_case_wgs_single"
    option_key = "--priority"
    option_value = "high"
    # WHEN ensuring case config exists where it should be stored
    Path.mkdir(
        Path(balsamic_context["BalsamicAnalysisAPI"].get_config_path(case_id)).parent, exist_ok=True
    )
    Path(balsamic_context["BalsamicAnalysisAPI"].get_config_path(case_id)).touch(exist_ok=True)
    # WHEN dry running with option specified
    result = cli_runner.invoke(
        run, [case_id, "--dry-run", option_key, option_value], obj=balsamic_context
    )
    # THEN command should execute successfully
    assert result.exit_code == EXIT_SUCCESS
    # THEN dry-print should include the the option-value
    assert "--qos" in caplog.text
    assert option_value in caplog.text


def test_priority_clinical(cli_runner, balsamic_context: dict, caplog):
    """Test command with case_id set to default NORMAL priority, when priority is not set manually"""
    caplog.set_level(logging.INFO)
    # GIVEN valid case-id
    case_id = "balsamic_case_wgs_single"
    option_value = "normal"
    # WHEN ensuring case config exists where it should be stored
    Path.mkdir(
        Path(balsamic_context["BalsamicAnalysisAPI"].get_config_path(case_id)).parent, exist_ok=True
    )
    Path(balsamic_context["BalsamicAnalysisAPI"].get_config_path(case_id)).touch(exist_ok=True)
    # WHEN dry running with option specified
    result = cli_runner.invoke(run, [case_id, "--dry-run"], obj=balsamic_context)
    # THEN command should execute successfully
    assert result.exit_code == EXIT_SUCCESS
    # THEN dry-print should include the the option-value
    assert "--qos" in caplog.text
    assert option_value in caplog.text



"""
This module handles concatenation of usalt fastq files.

Classes:
    FastqFileNameCreator: Creates valid usalt filenames
    FastqHandler: Handles fastq file linking
"""
import logging
from pathlib import Path
from typing import List

from cg.apps.pipelines.fastqhandler import BaseFastqHandler

LOGGER = logging.getLogger(__name__)


class FastqHandler(BaseFastqHandler):
    """Handles fastq file linking"""

    class FastqFileNameCreator(BaseFastqHandler.BaseFastqFileNameCreator):
        """Creates valid usalt filename from the parameters"""

        @staticmethod
        def create(lane: str, flowcell: str, sample: str, read: str, more: dict = None) -> str:
            """Name a FASTQ file following usalt conventions. Naming must be
            xxx_R_1.fastq.gz and xxx_R_2.fastq.gz"""

            undetermined = more.get("undetermined", None)

            # ACC1234A1_FCAB1ABC2_L1_1.fastq.gz sample_flowcell_lane_read.fastq.gz
            flowcell = f"{flowcell}-undetermined" if undetermined else flowcell
            return f"{sample}_{flowcell}_L{lane}_{read}.fastq.gz"

    def __init__(self, config):
        super().__init__(config)
        self.root_dir = config["microsalt"]["root"]

    def link(self, ticket: int, sample: str, files: List[dict]) -> None:
        """Link FASTQ files for a usalt sample.
        Shall be linked to /<usalt root directory>/ticket/fastq/"""

        # The fastq files should be linked to /.../fastq/<ticket>/<sample>/*.fastq.gz.
        wrk_dir = Path(self.root_dir) / "fastq" / str(ticket) / sample

        wrk_dir.mkdir(parents=True, exist_ok=True)

        for fastq_data in files:
            original_fastq_path = Path(fastq_data["path"])
            linked_fastq_name = self.FastqFileNameCreator.create(
                lane=fastq_data["lane"],
                flowcell=fastq_data["flowcell"],
                sample=sample,
                read=fastq_data["read"],
                more={"undetermined": fastq_data["undetermined"]},
            )
            linked_fastq_path = wrk_dir / linked_fastq_name

            if not linked_fastq_path.exists():
                LOGGER.info("linking: %s -> %s", original_fastq_path, linked_fastq_path)
                linked_fastq_path.symlink_to(original_fastq_path)
            else:
                LOGGER.debug("destination path already exists: %s", linked_fastq_path)



"""API for transfer a flowcell"""

import logging
from typing import List

from cg.apps.hk import HousekeeperAPI
from cg.apps.stats import StatsAPI
from cg.store import Store, models

LOG = logging.getLogger(__name__)


class TransferFlowcell:
    """Transfer flowcell API"""

    def __init__(self, db: Store, stats_api: StatsAPI, hk_api: HousekeeperAPI = None):
        self.db = db
        self.stats = stats_api
        self.hk = hk_api

    def transfer(self, flowcell_name: str, store: bool = True) -> models.Flowcell:
        """Populate the database with the information."""
        if store and self.hk.tag("fastq") is None:
            self.hk.add_commit(self.hk.new_tag("fastq"))
        if store and self.hk.tag(flowcell_name) is None:
            self.hk.add_commit(self.hk.new_tag(flowcell_name))
        stats_data = self.stats.flowcell(flowcell_name)
        flowcell_obj = self.db.flowcell(flowcell_name)
        if flowcell_obj is None:
            flowcell_obj = self.db.add_flowcell(
                name=flowcell_name,
                sequencer=stats_data["sequencer"],
                sequencer_type=stats_data["sequencer_type"],
                date=stats_data["date"],
            )
        flowcell_obj.status = "ondisk"
        for sample_data in stats_data["samples"]:
            LOG.debug(f"adding reads/fastqs to sample: {sample_data['name']}")
            sample_obj = self.db.sample(sample_data["name"])
            if sample_obj is None:
                LOG.warning(f"unable to find sample: {sample_data['name']}")
                continue

            if store:
                self.store_fastqs(
                    sample=sample_obj.internal_id,
                    flowcell=flowcell_name,
                    fastq_files=sample_data["fastqs"],
                )

            sample_obj.reads = sample_data["reads"]
            enough_reads = (
                sample_obj.reads > sample_obj.application_version.application.expected_reads
            )
            newest_date = (sample_obj.sequenced_at is None) or (
                flowcell_obj.sequenced_at > sample_obj.sequenced_at
            )
            if enough_reads and newest_date:
                sample_obj.sequenced_at = flowcell_obj.sequenced_at

            if isinstance(sample_obj, models.Sample):
                flowcell_obj.samples.append(sample_obj)

            LOG.info(
                f"added reads to sample: {sample_data['name']} - {sample_data['reads']} "
                f"[{'DONE' if enough_reads else 'NOT DONE'}]"
            )

        return flowcell_obj

    def store_fastqs(self, sample: str, flowcell: str, fastq_files: List[str]):
        """Store FASTQ files for a sample in housekeeper."""
        hk_bundle = self.hk.bundle(sample)
        if hk_bundle is None:
            hk_bundle = self.hk.new_bundle(sample)
            self.hk.add_commit(hk_bundle)
            new_version = self.hk.new_version(created_at=hk_bundle.created_at)
            hk_bundle.versions.append(new_version)
            self.hk.commit()
            LOG.info(f"added new Housekeeper bundle: {hk_bundle.name}")

        with self.hk.session_no_autoflush():
            hk_version = hk_bundle.versions[0]
            for fastq_file in fastq_files:
                if self.hk.files(path=fastq_file).first() is None:
                    LOG.info(f"found FASTQ file: {fastq_file}")
                    tags = [self.hk.tag("fastq"), self.hk.tag(flowcell)]
                    new_file = self.hk.new_file(path=fastq_file, tags=tags)
                    hk_version.files.append(new_file)
            self.hk.commit()



"""Fixtures for testing the madeline cg app"""

from pathlib import Path

import pytest

from cg.apps.madeline.api import MadelineAPI


@pytest.fixture(name="madeline_output")
def fixture_madeline_output():
    """File with madeline output"""
    return "tests/fixtures/apps/madeline/madeline.xml"


@pytest.fixture
def madeline_columns():
    """return a dictionary with madeline columns"""
    columns = {
        "family": "FamilyId",
        "sample": "IndividualId",
        "sex": "Gender",
        "father": "Father",
        "mother": "Mother",
        "deceased": "Deceased",
        "proband": "Proband",
        "status": "Affected",
    }
    return columns


@pytest.fixture(name="mother")
def fixture_mother():
    """return a dictionary with ind info"""
    ind_info = {
        "sample": "mother",
        "sex": "female",
        "proband": False,
        "status": "unaffected",
    }
    return ind_info


@pytest.fixture(name="father")
def fixture_father():
    """return a dictionary with ind info"""
    ind_info = {
        "sample": "father",
        "sex": "male",
        "proband": False,
        "status": "unaffected",
    }
    return ind_info


@pytest.fixture(name="proband")
def fixture_proband():
    """return a dictionary with ind info"""
    ind_info = {
        "sample": "proband",
        "sex": "female",
        "father": "father",
        "mother": "mother",
        "deceased": False,
        "proband": True,
        "status": "affected",
    }
    return ind_info


@pytest.fixture(name="trio")
def fixture_trio(proband, mother, father):
    """return a list with a trio"""
    return [proband, mother, father]


@pytest.fixture
def madeline_input(proband):
    """return a iterable with madeline formated lines"""
    inds = [proband]
    case_id = "test"
    _input = []
    for line in MadelineAPI.make_ped(case_id, inds):
        _input.append(line)

    return _input


@pytest.fixture
def madeline_process(madeline_output):
    """return a madeline process mock"""
    return MadelineProcessMock(binary="madeline", madeline_file=madeline_output)


class MadelineProcessMock:
    """Class to mock simple madeline process calls"""

    def __init__(self, binary, config=None, config_parameter="--config", madeline_file=None):
        """Initialise mock"""
        self.binary = binary
        self.config = config
        self.config_parameter = config_parameter
        self.stdout = ""
        self._madeline_content = ""
        if madeline_file:
            with open(madeline_file, "r") as output:
                self._madeline_content = output.read()

    def run_command(self, parameters):
        """Mock that a command is run"""
        output_prefix = Path(parameters[-2])
        outfile = output_prefix.with_suffix(".xml")
        with open(outfile, "w") as file_handle:
            file_handle.write(self._madeline_content)

        return outfile

    def stdout_lines(self):
        """Iterate over the lines in self.stdout"""
        for line in self.stdout.split("\n"):
            yield line



""" Start of CLI """
import logging
import sys

import click
import coloredlogs
import ruamel.yaml

import cg
from cg.store import Store

from .add import add
from .backup import backup
from .clean import clean
from .compress.base import compress, decompress
from .export import export
from .get import get
from .import_cmd import import_cmd
from .reset import reset_cmd
from cg.cli.set.base import set_cmd
from .status import status
from .store import store as store_cmd
from .transfer import transfer
from .upload import vogue as vogue_command
from .upload.base import upload
from .workflow.base import workflow as workflow_cmd
from .deploy.base import deploy as deploy_cmd

LOG = logging.getLogger(__name__)
LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR"]


@click.group()
@click.option("-c", "--config", type=click.File(), help="path to config file")
@click.option("-d", "--database", help="path/URI of the SQL database")
@click.option(
    "-l", "--log-level", type=click.Choice(LEVELS), default="INFO", help="lowest level to log at"
)
@click.version_option(cg.__version__, prog_name=cg.__title__)
@click.pass_context
def base(context, config, database, log_level):
    """cg - interface between tools at Clinical Genomics."""
    log_format = "%(message)s" if sys.stdout.isatty() else None
    coloredlogs.install(level=log_level, fmt=log_format)
    context.obj = ruamel.yaml.safe_load(config) if config else {}
    if database:
        context.obj["database"] = database


@base.command()
@click.option("--reset", is_flag=True, help="reset database before setting up tables")
@click.option("--force", is_flag=True, help="bypass manual confirmations")
@click.pass_context
def init(context, reset, force):
    """Setup the database."""
    status_db = Store(context.obj["database"])
    existing_tables = status_db.engine.table_names()
    if force or reset:
        if existing_tables and not force:
            message = f"Delete existing tables? [{', '.join(existing_tables)}]"
            click.confirm(click.style(message, fg="yellow"), abort=True)
        status_db.drop_all()
    elif existing_tables:
        LOG.error("Database already exists, use '--reset'")
        context.abort()

    status_db.create_all()
    LOG.info("Success! New tables: %s", ", ".join(status_db.engine.table_names()))


base.add_command(add)
base.add_command(backup)
base.add_command(clean)
base.add_command(compress)
base.add_command(decompress)
base.add_command(export)
base.add_command(get)
base.add_command(import_cmd)
base.add_command(reset_cmd)
base.add_command(set_cmd)
base.add_command(status)
base.add_command(transfer)
base.add_command(upload)
base.add_command(workflow_cmd)
base.add_command(store_cmd)
base.add_command(deploy_cmd)

upload.add_command(vogue_command)



"""Tests for the upload genotypes api"""

from pathlib import Path
from datetime import datetime

from cg.meta.upload.gt import UploadGenotypesAPI
from cg.apps import hk
from cg.store import models


def test_get_sample_sex():
    """Test to get the predicted sex from a MIP run using the upload genotypes API"""
    # GIVEN a UploadGenotypesAPI some qcmetrics data
    qcmetrics_parsed = {
        "samples": [
            {
                "id": "father",
                "predicted_sex": "male",
            },
            {
                "id": "child",
                "predicted_sex": "male",
            },
            {
                "id": "mother",
                "predicted_sex": "female",
            },
        ]
    }

    # WHEN fetching the predicted gender for the individual with id "father"
    sex = UploadGenotypesAPI.get_sample_predicted_sex(
        sample_id="father", parsed_qcmetrics_data=qcmetrics_parsed
    )

    # THEN assert that the gender was "male"
    assert sex == "male"


def test_get_parsed_qc_metrics_data(case_qc_metrics: Path):
    """Test to get the predicted sex from a MIP run using the upload genotypes API"""
    # GIVEN a UploadGenotypesAPI and the path to a qcmetrics file with case data

    # WHEN fetching the predicted gender for the individual with id "father"

    qcmetrics_parsed = UploadGenotypesAPI.get_parsed_qc_metrics_data(case_qc_metrics)

    # THEN assert that there are three samples in the case qc metrics file
    assert len(qcmetrics_parsed["samples"]) == 3


def test_get_bcf_file(upload_genotypes_api: UploadGenotypesAPI, case_id: str, timestamp: datetime):
    """Test to get the predicted sex from a MIP run using the upload genotypes API"""
    # GIVEN a UploadGenotypesAPI populated with some data in housekeeper
    hk_version = upload_genotypes_api.hk.version(case_id, timestamp)

    # WHEN fetching the gbcf file with the api
    gbcf = upload_genotypes_api.get_bcf_file(hk_version)

    # THEN assert that the file has the correct tag
    assert "snv-gbcf" in (tag.name for tag in gbcf.tags)


def test_get_data(upload_genotypes_api: UploadGenotypesAPI, analysis_obj: models.Analysis):
    """Test to get data from the UploadGenotypesAPI"""
    # GIVEN a UploadGenotypeAPI populated with some data
    # GIVEN an analysis object with a trio

    # WHEN parsing the data
    result = upload_genotypes_api.data(analysis_obj=analysis_obj)

    # THEN assert that the result looks like expected
    assert len(result["samples_sex"]) == 3



"""
    Module for loqusdb API
"""

import json
import logging
from subprocess import CalledProcessError

from cg.exc import CaseNotFoundError
from cg.utils import Process

LOG = logging.getLogger(__name__)


class LoqusdbAPI:

    """
    API for loqusdb
    """

    def __init__(self, config: dict, analysis_type: str = "wgs"):
        super(LoqusdbAPI, self).__init__()

        self.analysis_type = analysis_type

        self.loqusdb_config = config["loqusdb"]["config_path"]
        self.loqusdb_binary = config["loqusdb"]["binary_path"]

        if self.analysis_type == "wes":
            self.loqusdb_config = config["loqusdb-wes"]["config_path"]
            self.loqusdb_binary = config["loqusdb-wes"]["binary_path"]

        self.process = Process(self.loqusdb_binary, self.loqusdb_config)

    def load(
        self, family_id: str, ped_path: str, vcf_path: str, gbcf_path: str, vcf_sv_path: str = None
    ) -> dict:
        """Add observations from a VCF."""
        load_call_parameters = [
            "load",
            "-c",
            family_id,
            "-f",
            ped_path,
            "--variant-file",
            vcf_path,
            "--check-profile",
            gbcf_path,
            "--hard-threshold",
            "0.95",
            "--soft-threshold",
            "0.90",
        ]
        if self.analysis_type == "wgs" and vcf_sv_path:
            load_call_parameters.extend(["--sv-variants", vcf_sv_path])

        nr_variants = 0
        self.process.run_command(parameters=load_call_parameters)
        for line in self.process.stderr_lines():
            line_content = line.split("INFO")[-1].strip()
            if "inserted" in line_content:
                nr_variants = int(line_content.split(":")[-1].strip())

        return dict(variants=nr_variants)

    def get_case(self, case_id: str) -> dict:
        """Find a case in the database by case id."""
        case_obj = None
        cases_parameters = ["cases", "-c", case_id, "--to-json"]

        self.process.run_command(parameters=cases_parameters)

        output = self.process.stdout

        # If case not in loqusdb, stdout of loqusdb command will be empty.
        if not output:
            raise CaseNotFoundError(f"Case {case_id} not found in loqusdb")

        case_obj = json.loads(output)[0]

        return case_obj

    def get_duplicate(self, vcf_file: str) -> dict:
        """Find matching profiles in loqusdb"""
        ind_obj = {}
        duplicates_params = ["profile", "--check-vcf", vcf_file, "--profile-threshold", "0.95"]

        try:
            self.process.run_command(parameters=duplicates_params)
        except CalledProcessError:
            # If CalledProcessError is raised, log and raise error
            LOG.critical("Could not run profile command")
            raise

        output = self.process.stdout

        if not output:
            LOG.info("No duplicates found")
            return ind_obj

        ind_obj = json.loads(output)

        return ind_obj

    def __repr__(self):

        return f"LoqusdbAPI(binary={self.loqusdb_binary}," f"config={self.loqusdb_config})"



"""CLI support to create config and/or start BALSAMIC """

import logging

import click

from cg.apps.balsamic.api import BalsamicAPI
from cg.apps.balsamic.fastq import FastqHandler
from cg.apps.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.cli.workflow.balsamic.deliver import deliver as deliver_cmd
from cg.exc import BalsamicStartError, BundleAlreadyAddedError, LimsDataError
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.store import Store
from cg.constants import EXIT_SUCCESS, EXIT_FAIL

LOG = logging.getLogger(__name__)

ARGUMENT_CASE_ID = click.argument("case_id", required=True)
OPTION_DRY = click.option(
    "-d", "--dry-run", "dry", help="Print command to console without executing", is_flag=True
)
OPTION_PANEL_BED = click.option(
    "--panel-bed",
    required=False,
    help="Panel BED is determined based on capture kit \
    used for library prep. Set this option to override the default",
)
OPTION_ANALYSIS_TYPE = click.option(
    "-a",
    "--analysis-type",
    type=click.Choice(["qc", "paired", "single"]),
    help="Setting this option to qc ensures only QC analysis is performed",
)
OPTION_RUN_ANALYSIS = click.option(
    "-r",
    "--run-analysis",
    is_flag=True,
    default=False,
    help="Execute BALSAMIC in non-dry mode",
)
OPTION_PRIORITY = click.option(
    "-p",
    "--priority",
    type=click.Choice(["low", "normal", "high"]),
    help="Job priority in SLURM. Will be set automatically according to priority in ClinicalDB, \
         this option can be used to override server setting",
)


@click.group(invoke_without_command=True)
@OPTION_ANALYSIS_TYPE
@OPTION_PRIORITY
@OPTION_PANEL_BED
@OPTION_RUN_ANALYSIS
@OPTION_DRY
@click.pass_context
def balsamic(context, priority, panel_bed, analysis_type, run_analysis, dry):
    """Cancer analysis workflow """
    if context.invoked_subcommand is None:
        click.echo(context.get_help())
        return None
    config = context.obj
    context.obj["BalsamicAnalysisAPI"] = BalsamicAnalysisAPI(
        balsamic_api=BalsamicAPI(config),
        store=Store(config["database"]),
        housekeeper_api=HousekeeperAPI(config),
        fastq_handler=FastqHandler(config),
        lims_api=LimsAPI(config),
    )


@balsamic.command("link")
@ARGUMENT_CASE_ID
@click.pass_context
def link(context, case_id):
    """
    Locates FASTQ files for given CASE_ID.
    The files are renamed, concatenated, and saved in BALSAMIC working directory
    """
    balsamic_analysis_api = context.obj["BalsamicAnalysisAPI"]
    try:
        LOG.info(f"Linking samples in case {case_id}")
        balsamic_analysis_api.link_samples(case_id)
    except BalsamicStartError as e:
        LOG.error(f"Could not link samples: {e.message}")
        raise click.Abort()


@balsamic.command("config-case")
@ARGUMENT_CASE_ID
@OPTION_PANEL_BED
@OPTION_DRY
@click.pass_context
def config_case(context, panel_bed, case_id, dry):
    """Create config file for BALSAMIC analysis for a given CASE_ID"""
    balsamic_analysis_api = context.obj["BalsamicAnalysisAPI"]

    try:
        LOG.info(f"Creating config file for {case_id}.")
        arguments = balsamic_analysis_api.get_verified_config_case_arguments(
            case_id=case_id,
            panel_bed=panel_bed,
        )
        balsamic_analysis_api.balsamic_api.config_case(arguments=arguments, dry=dry)
    except (BalsamicStartError, LimsDataError) as e:
        LOG.error(f"Could not create config: {e.message}")
        raise click.Abort()


@balsamic.command("run")
@ARGUMENT_CASE_ID
@OPTION_DRY
@OPTION_PRIORITY
@OPTION_ANALYSIS_TYPE
@OPTION_RUN_ANALYSIS
@click.pass_context
def run(context, analysis_type, run_analysis, priority, case_id, dry):
    """Run balsamic analysis for given CASE ID"""
    balsamic_analysis_api = context.obj["BalsamicAnalysisAPI"]
    try:
        LOG.info(f"Running analysis for {case_id}")
        arguments = {
            "priority": priority or balsamic_analysis_api.get_priority(case_id),
            "analysis_type": analysis_type,
            "run_analysis": run_analysis,
            "sample_config": balsamic_analysis_api.get_config_path(
                case_id=case_id, check_exists=True
            ),
        }
        balsamic_analysis_api.balsamic_api.run_analysis(
            arguments=arguments, run_analysis=run_analysis, dry=dry
        )
        balsamic_analysis_api.set_statusdb_action(case_id=case_id, action="running")
    except BalsamicStartError as e:
        LOG.error(f"Could not run analysis: {e.message}")
        raise click.Abort()


@balsamic.command("report-deliver")
@ARGUMENT_CASE_ID
@OPTION_DRY
@OPTION_ANALYSIS_TYPE
@click.pass_context
def report_deliver(context, case_id, analysis_type, dry):
    """Create a housekeeper deliverables file for given CASE ID"""
    balsamic_analysis_api = context.obj["BalsamicAnalysisAPI"]
    try:
        LOG.info(f"Creating delivery report for {case_id}")
        case_object = balsamic_analysis_api.get_case_object(case_id)
        sample_config = balsamic_analysis_api.get_config_path(case_id=case_id, check_exists=True)
        analysis_finish = balsamic_analysis_api.get_analysis_finish_path(case_id, check_exists=True)
        LOG.info(f"Found analysis finish file: {analysis_finish}")
        arguments = {"sample_config": sample_config, "analysis_type": analysis_type}
        balsamic_analysis_api.balsamic_api.report_deliver(arguments=arguments, dry=dry)
    except BalsamicStartError as e:
        LOG.error(f"Could not create report file: {e.message}")
        raise click.Abort()


@balsamic.command("store-housekeeper")
@ARGUMENT_CASE_ID
@click.pass_context
def store_housekeeper(context, case_id):
    """Store a finished analysis in Housekeeper and StatusDB."""
    balsamic_analysis_api = context.obj["BalsamicAnalysisAPI"]
    try:
        LOG.info(f"Storing bundle data in Housekeeper for {case_id}")
        balsamic_analysis_api.upload_bundle_housekeeper(case_id=case_id)
        LOG.info(f"Storing Analysis in ClinicalDB for {case_id}")
        balsamic_analysis_api.upload_analysis_statusdb(case_id=case_id)
    except (BundleAlreadyAddedError, FileExistsError) as e:
        LOG.error(f"Could not store bundle in Housekeeper and StatusDB: {e.message}!")
        balsamic_analysis_api.housekeeper_api.rollback()
        balsamic_analysis_api.store.rollback()
        raise click.Abort()
    except BalsamicStartError as e:
        LOG.error(f"Could not store bundle in Housekeeper and StatusDB: {e.message}!")
        raise click.Abort()


@balsamic.command("start")
@ARGUMENT_CASE_ID
@OPTION_ANALYSIS_TYPE
@OPTION_PRIORITY
@OPTION_DRY
@OPTION_PANEL_BED
@click.pass_context
def start(context, case_id, analysis_type, panel_bed, priority, dry):
    """Start full workflow for CASE ID"""
    LOG.info(f"Starting analysis for {case_id}")
    context.invoke(link, case_id=case_id)
    context.invoke(config_case, case_id=case_id, panel_bed=panel_bed, dry=dry)
    context.invoke(
        run,
        case_id=case_id,
        analysis_type=analysis_type,
        priority=priority,
        run_analysis="--run-analysis",
        dry=dry,
    )


@balsamic.command("store")
@ARGUMENT_CASE_ID
@OPTION_DRY
@OPTION_ANALYSIS_TYPE
@click.pass_context
def store(context, case_id, analysis_type, dry):
    """Generate Housekeeper report for CASE ID and store in Housekeeper"""
    LOG.info(f"Storing analysis for {case_id}")
    context.invoke(report_deliver, case_id=case_id, analysis_type=analysis_type, dry=dry)
    context.invoke(store_housekeeper, case_id=case_id)


@balsamic.command("start-available")
@OPTION_DRY
@click.pass_context
def start_available(context, dry):
    """Start full workflow for all available BALSAMIC cases"""
    balsamic_analysis_api = context.obj["BalsamicAnalysisAPI"]
    exit_code = EXIT_SUCCESS
    for case_id in balsamic_analysis_api.get_cases_to_analyze():
        try:
            context.invoke(start, case_id=case_id, dry=dry)
        except click.Abort:
            exit_code = EXIT_FAIL
        except Exception as e:
            LOG.error(f"Unspecified error occurred - {e}")
            exit_code = EXIT_FAIL
    if exit_code:
        raise click.Abort()


@balsamic.command("store-available")
@OPTION_DRY
@click.pass_context
def store_available(context, dry):
    """Store bundle data for all available Balsamic cases"""
    balsamic_analysis_api = context.obj["BalsamicAnalysisAPI"]
    exit_code = EXIT_SUCCESS
    for case_id in balsamic_analysis_api.get_cases_to_store():
        try:
            context.invoke(store, case_id=case_id, dry=dry)
        except click.Abort:
            exit_code = EXIT_FAIL
        except Exception as e:
            LOG.error(f"Unspecified error occurred - {e}")
            exit_code = EXIT_FAIL
    if exit_code:
        raise click.Abort()


balsamic.add_command(deliver_cmd)



"""Test the cli for uploading using auto"""
import logging

from cg.cli.upload.base import auto


def test_upload_auto_with_pipeline_as_argument(
    base_context, cli_runner, caplog, helpers, sample_store, timestamp
):
    """Test upload auto"""
    # GIVEN a store with a MIP analysis
    helpers.add_analysis(store=sample_store, completed_at=timestamp, pipeline="MIP")

    # WHEN uploading all analysis from pipeline MIP
    caplog.set_level(logging.INFO)
    cli_runner.invoke(auto, ["--pipeline", "MIP"], obj=base_context)

    # THEN assert that the MIP analysis was successfully uploaded
    with caplog.at_level(logging.INFO):
        assert "uploading family" in caplog.text



"""Fixtures for cli clean tests"""

import datetime as dt
from pathlib import Path

import pytest

from cg.apps.balsamic.api import BalsamicAPI
from cg.apps.hk import HousekeeperAPI
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.store import Store


@pytest.fixture
def balsamic_clean_store(base_store: Store, timestamp_yesterday: dt.datetime, helpers) -> Store:
    store = base_store

    # Create textbook case for cleaning
    case_to_clean = helpers.add_family(
        store=store, internal_id="balsamic_case_clean", family_id="balsamic_case_clean"
    )
    sample_case_to_clean = helpers.add_sample(
        store,
        internal_id="balsamic_sample_clean",
        is_tumour=True,
        application_type="wgs",
        data_analysis="balsamic",
    )
    helpers.add_relationship(store, family=case_to_clean, sample=sample_case_to_clean)

    helpers.add_analysis(
        store,
        family=case_to_clean,
        pipeline="balsamic",
        started_at=timestamp_yesterday,
        uploaded_at=timestamp_yesterday,
        cleaned_at=None,
    )
    return store


@pytest.fixture(name="balsamic_dir")
def balsamic_dir(tmpdir_factory, apps_dir: Path) -> Path:
    """Return the path to the balsamic apps dir"""
    balsamic_dir = tmpdir_factory.mktemp("balsamic")
    return Path(balsamic_dir).absolute().as_posix()


@pytest.fixture
def server_config(balsamic_dir: Path) -> dict:
    return {
        "database": "database",
        "bed_path": balsamic_dir,
        "balsamic": {
            "root": balsamic_dir,
            "singularity": Path(balsamic_dir, "singularity.sif").as_posix(),
            "reference_config": Path(balsamic_dir, "reference.json").as_posix(),
            "binary_path": "/home/proj/bin/conda/envs/S_BALSAMIC/bin/balsamic",
            "conda_env": "S_BALSAMIC",
            "slurm": {
                "mail_user": "test.mail@scilifelab.se",
                "account": "development",
                "qos": "low",
            },
        },
        "housekeeper": {
            "database": "database",
            "root": balsamic_dir,
        },
        "lims": {
            "host": "example.db",
            "username": "testuser",
            "password": "testpassword",
        },
    }


@pytest.fixture
def balsamic_analysis_api(
    server_config: dict, balsamic_clean_store: Store, housekeeper_api: HousekeeperAPI
):
    return BalsamicAnalysisAPI(
        balsamic_api=BalsamicAPI(server_config),
        store=balsamic_clean_store,
        housekeeper_api=housekeeper_api,
        fastq_handler="FastqHandler",
        lims_api="LIMS",
        fastq_api="FastqAPI",
    )


@pytest.fixture
def clean_context(
    base_store: Store,
    housekeeper_api: HousekeeperAPI,
    balsamic_analysis_api: BalsamicAnalysisAPI,
    helpers,
    tmpdir,
) -> dict:
    """context to use in cli"""

    return {
        "hk_api": housekeeper_api,
        "store_api": balsamic_clean_store,
        "BalsamicAnalysisAPI": balsamic_analysis_api,
    }


@pytest.fixture
def balsamic_analysis_api(
    server_config: dict, balsamic_clean_store: Store, housekeeper_api: HousekeeperAPI
):
    return BalsamicAnalysisAPI(
        balsamic_api=BalsamicAPI(server_config),
        store=balsamic_clean_store,
        housekeeper_api=housekeeper_api,
        fastq_handler="FastqHandler",
        lims_api="LIMS",
    )


@pytest.fixture
def clean_context(
    base_store: Store,
    housekeeper_api: HousekeeperAPI,
    balsamic_analysis_api: BalsamicAnalysisAPI,
    helpers,
    tmpdir,
) -> dict:
    """context to use in cli"""

    return {
        "hk_api": housekeeper_api,
        "store_api": balsamic_clean_store,
        "BalsamicAnalysisAPI": balsamic_analysis_api,
    }



import logging

from typing import List, Dict

from cg.utils.commands import Process

LOG = logging.getLogger(__name__)


class BalsamicAPI:
    """Handles execution of BALSAMIC"""

    __EXIT_SUCCESS = 0

    def __init__(self, config: dict):
        self.binary = config["balsamic"]["binary_path"]
        self.singularity = config["balsamic"]["singularity"]
        self.reference_config = config["balsamic"]["reference_config"]
        self.root_dir = config["balsamic"]["root"]
        self.account = config["balsamic"]["slurm"]["account"]
        self.email = config["balsamic"]["slurm"]["mail_user"]
        self.qos = config["balsamic"]["slurm"]["qos"]
        self.bed_path = config["bed_path"]
        self.process = Process(self.binary)

    def __build_command_str(self, options: dict) -> List[str]:
        formatted_options = []
        for key, val in options.items():
            if val:
                formatted_options.append(str(key))
                formatted_options.append(str(val))
        return formatted_options

    def config_case(self, arguments: dict, dry: bool = False) -> int:
        """Create config file for BALSAMIC analysis """

        command = ["config", "case"]
        options = self.__build_command_str(
            {
                "--analysis-dir": self.root_dir,
                "--singularity": self.singularity,
                "--reference-config": self.reference_config,
                "--case-id": arguments.get("case_id"),
                "--normal": arguments.get("normal"),
                "--tumor": arguments.get("tumor"),
                "--panel-bed": arguments.get("panel_bed"),
                "--umi-trim-length": arguments.get("umi_trim_length"),
            }
        )
        parameters = command + options
        if dry:
            LOG.info(f'Dry run command balsamic {" ".join(parameters)}')
            retcode = self.__EXIT_SUCCESS
        else:
            retcode = self.process.run_command(parameters=parameters)
        return retcode

    def run_analysis(self, arguments: dict, run_analysis: bool, dry: bool = False) -> int:
        """Execute BALSAMIC run analysis with given options"""

        command = ["run", "analysis"]
        run_analysis = ["--run-analysis"] if run_analysis else []
        options = self.__build_command_str(
            {
                "--account": self.account,
                "--mail-user": arguments.get("email", self.email),
                "--qos": arguments.get("priority", self.qos),
                "--sample-config": arguments.get("sample_config"),
                "--analysis-type": arguments.get("analysis_type"),
            }
        )
        parameters = command + options + run_analysis
        if dry:
            LOG.info(f'Dry run command balsamic {" ".join(parameters)}')
            retcode = self.__EXIT_SUCCESS
        else:
            retcode = self.process.run_command(parameters=parameters)
        return retcode

    def report_deliver(self, arguments: dict, dry: bool = False) -> int:
        """Execute BALSAMIC report deliver with given options"""

        command = ["report", "deliver"]
        options = self.__build_command_str(
            {
                "--sample-config": arguments.get("sample_config"),
                "--analysis-type": arguments.get("analysis_type"),
            }
        )
        parameters = command + options
        if dry:
            LOG.info(f'Dry run command balsamic {" ".join(parameters)}')
            retcode = self.__EXIT_SUCCESS
        else:
            retcode = self.process.run_command(parameters=parameters)
        return retcode



from pathlib import Path
import json
import logging
import pytest

from cg.cli.workflow.balsamic.base import (
    balsamic,
    start,
    store,
    start_available,
    store_available,
)

EXIT_SUCCESS = 0


def test_balsamic_no_args(cli_runner, balsamic_context: dict):
    """Test to see that running BALSAMIC without options prints help and doesn't result in an error"""
    # GIVEN no arguments or options besides the command call

    # WHEN running command
    result = cli_runner.invoke(balsamic, [], obj=balsamic_context)

    # THEN command runs successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN help should be printed
    assert "help" in result.output


def test_start(cli_runner, balsamic_context: dict, mock_config, caplog):
    """Test to ensure all parts of start command will run successfully given ideal conditions"""
    caplog.set_level(logging.INFO)

    # GIVEN case id for which we created a config file
    case_id = "balsamic_case_wgs_single"

    # WHEN dry running with dry specified
    result = cli_runner.invoke(start, [case_id, "--dry-run"], obj=balsamic_context)

    # THEN command should execute successfully
    assert result.exit_code == EXIT_SUCCESS
    assert case_id in caplog.text


def test_store(
    cli_runner,
    balsamic_context: dict,
    real_housekeeper_api,
    mock_config,
    mock_deliverable,
    mock_analysis_finish,
    caplog,
):
    """Test to ensure all parts of store command are run successfully given ideal conditions"""
    caplog.set_level(logging.INFO)

    # GIVEN case-id for which we created a config file, deliverables file, and analysis_finish file
    case_id = "balsamic_case_wgs_single"

    # Set Housekeeper to an empty real Housekeeper store
    balsamic_context["BalsamicAnalysisAPI"].housekeeper_api = real_housekeeper_api

    # Make sure the bundle was not present in the store
    assert not balsamic_context["BalsamicAnalysisAPI"].housekeeper_api.bundle(case_id)

    # Make sure  analysis not already stored in ClinicalDB
    assert not balsamic_context["BalsamicAnalysisAPI"].store.family(case_id).analyses

    # WHEN running command
    result = cli_runner.invoke(store, [case_id, "--dry-run"], obj=balsamic_context)

    # THEN bundle should be successfully added to HK and STATUS
    assert result.exit_code == EXIT_SUCCESS
    assert "Analysis successfully stored in Housekeeper" in caplog.text
    assert "Analysis successfully stored in ClinicalDB" in caplog.text
    assert balsamic_context["BalsamicAnalysisAPI"].store.family(case_id).analyses
    assert balsamic_context["BalsamicAnalysisAPI"].housekeeper_api.bundle(case_id)


def test_start_available(tmpdir_factory, cli_runner, balsamic_context: dict, caplog):
    """Test to ensure all parts of compound start-available command are executed given ideal conditions
    Test that start-available picks up eligible cases and does not pick up ineligible ones"""
    caplog.set_level(logging.INFO)

    # GIVEN CASE ID of sample where read counts pass threshold
    case_id_success = "balsamic_case_wgs_paired_enough_reads"

    # GIVEN CASE ID where read counts did not pass the threshold
    case_id_fail = "balsamic_case_tgs_paired"

    # Ensure the config is mocked to run compound command
    Path.mkdir(
        Path(balsamic_context["BalsamicAnalysisAPI"].get_config_path(case_id_success)).parent,
        exist_ok=True,
    )
    Path(balsamic_context["BalsamicAnalysisAPI"].get_config_path(case_id_success)).touch(
        exist_ok=True
    )
    # WHEN running command
    result = cli_runner.invoke(start_available, ["--dry-run"], obj=balsamic_context)

    # THEN command exits with 1 because one of cases raised errors
    assert result.exit_code == 1

    # THEN it should successfully identify the one case eligible for auto-start
    assert case_id_success in caplog.text

    # THEN action of the case should be set to running
    assert balsamic_context["BalsamicAnalysisAPI"].store.family(case_id_success).action == "running"

    # THEN the ineligible case should NOT be ran
    assert case_id_fail not in caplog.text

    # THEN action of the case should NOT be set to running
    assert balsamic_context["BalsamicAnalysisAPI"].store.family(case_id_fail).action == None


def test_store_available(
    tmpdir_factory,
    cli_runner,
    balsamic_context: dict,
    real_housekeeper_api,
    mock_config,
    mock_deliverable,
    mock_analysis_finish,
    caplog,
):
    """Test to ensure all parts of compound store-available command are executed given ideal conditions
    Test that sore-available picks up eligible cases and does not pick up ineligible ones"""
    caplog.set_level(logging.INFO)

    # GIVEN CASE ID of sample where read counts pass threshold
    case_id_success = "balsamic_case_wgs_single"

    # GIVEN CASE ID where analysis finish is not mocked
    case_id_fail = "balsamic_case_wgs_paired"

    # Ensure the config is mocked for fail case to run compound command
    Path.mkdir(
        Path(balsamic_context["BalsamicAnalysisAPI"].get_config_path(case_id_fail)).parent,
        exist_ok=True,
    )
    Path(balsamic_context["BalsamicAnalysisAPI"].get_config_path(case_id_fail)).touch(exist_ok=True)

    # Ensure case was successfully picked up by start-available and status set to running
    result = cli_runner.invoke(start_available, ["--dry-run"], obj=balsamic_context)
    # THEN command exits with 1 because one of the cases threw errors
    assert result.exit_code == 1
    assert case_id_success in caplog.text
    assert balsamic_context["BalsamicAnalysisAPI"].store.family(case_id_success).action == "running"

    balsamic_context["BalsamicAnalysisAPI"].housekeeper_api = real_housekeeper_api
    # WHEN running command
    result = cli_runner.invoke(store_available, ["--dry-run"], obj=balsamic_context)

    # THEN command exits successfully
    assert result.exit_code == 0

    # THEN case id with analysis_finish gets picked up
    assert case_id_success in caplog.text

    # THEN bundle added successfully and action set to None
    assert balsamic_context["BalsamicAnalysisAPI"].store.family(case_id_success).action == None

    # THEN case id WITHOUT analysis_finish is still running
    assert balsamic_context["BalsamicAnalysisAPI"].store.family(case_id_fail).action == "running"

    # THEN case has analyses
    assert balsamic_context["BalsamicAnalysisAPI"].store.family(case_id_success).analyses

    # THEN bundle can be found in Housekeeper
    assert balsamic_context["BalsamicAnalysisAPI"].housekeeper_api.bundle(case_id_success)



import logging
import click

from cg.constants import PRIORITY_OPTIONS, STATUS_OPTIONS
from cg.store import Store

LOG = logging.getLogger(__name__)


@click.group()
@click.pass_context
def add(context):
    """Add new things to the database."""
    context.obj["db"] = Store(context.obj["database"])


@add.command()
@click.argument("internal_id")
@click.argument("name")
@click.option(
    "-cg",
    "--customer-group",
    "customer_group_id",
    required=False,
    help="internal ID for the customer group of the customer, a new group will be "
    "created if left out",
)
@click.option(
    "-ia",
    "--invoice-address",
    "invoice_address",
    required=True,
    help="Street adress, Post code, City",
)
@click.option(
    "-ir",
    "--invoice-reference",
    "invoice_reference",
    required=True,
    help="Invoice reference (text)",
)
@click.pass_context
def customer(
    context,
    internal_id: str,
    name: str,
    customer_group_id: str,
    invoice_address: str,
    invoice_reference: str,
):
    """Add a new customer with a unique INTERNAL_ID and NAME."""
    existing = context.obj["db"].customer(internal_id)
    if existing:
        LOG.error(f"{existing.name}: customer already added")
        context.abort()

    customer_group = context.obj["db"].customer_group(customer_group_id)
    if not customer_group:
        customer_group = context.obj["db"].add_customer_group(internal_id=internal_id, name=name)

    new_customer = context.obj["db"].add_customer(
        internal_id=internal_id,
        name=name,
        customer_group=customer_group,
        invoice_address=invoice_address,
        invoice_reference=invoice_reference,
    )
    context.obj["db"].add_commit(new_customer)
    message = f"customer added: {new_customer.internal_id} ({new_customer.id})"
    LOG.info(message)


@add.command()
@click.option("-a", "--admin", is_flag=True, help="make the user an admin")
@click.option(
    "-c",
    "--customer",
    "customer_id",
    required=True,
    help="internal ID for the customer of the user",
)
@click.argument("email")
@click.argument("name")
@click.pass_context
def user(context, admin, customer_id, email, name):
    """Add a new user with an EMAIL (login) and a NAME (full)."""
    customer_obj = context.obj["db"].customer(customer_id)
    existing = context.obj["db"].user(email)
    if existing:
        LOG.error(f"{existing.name}: user already added")
        context.abort()
    new_user = context.obj["db"].add_user(customer_obj, email, name, is_admin=admin)
    context.obj["db"].add_commit(new_user)
    LOG.info(f"user added: {new_user.email} ({new_user.id})")


@add.command()
@click.option("-l", "--lims", "lims_id", help="LIMS id for the sample")
@click.option("-d", "--downsampled", type=int, help="how many reads is the sample downsampled to?")
@click.option("-o", "--order", help="name of the order the sample belongs to")
@click.option(
    "-s",
    "--sex",
    type=click.Choice(["male", "female", "unknown"]),
    required=True,
    help="sample pedigree sex",
)
@click.option("-a", "--application", required=True, help="application tag name")
@click.option(
    "-p",
    "--priority",
    type=click.Choice(PRIORITY_OPTIONS),
    default="standard",
    help="set the priority for the samples",
)
@click.argument("customer_id")
@click.argument("name")
@click.pass_context
def sample(context, lims_id, downsampled, sex, order, application, priority, customer_id, name):
    """Add a sample for CUSTOMER_ID with a NAME (display)."""
    status = context.obj["db"]
    customer_obj = status.customer(customer_id)
    if customer_obj is None:
        LOG.error("customer not found")
        context.abort()
    application_obj = status.application(application)
    if application_obj is None:
        LOG.error("application not found")
        context.abort()
    new_record = status.add_sample(
        name=name,
        sex=sex,
        internal_id=lims_id,
        order=order,
        downsampled_to=downsampled,
        priority=priority,
    )
    new_record.application_version = status.current_application_version(application)
    new_record.customer = customer_obj
    status.add_commit(new_record)
    LOG.info(f"{new_record.internal_id}: new sample added")


@add.command()
@click.option(
    "--priority", type=click.Choice(PRIORITY_OPTIONS), default="standard", help="analysis priority"
)
@click.option("-p", "--panel", "panels", multiple=True, required=True, help="default gene panels")
@click.argument("customer_id")
@click.argument("name")
@click.pass_context
def family(context, priority, panels, customer_id, name):
    """Add a family to CUSTOMER_ID with a NAME."""
    status = context.obj["db"]
    customer_obj = status.customer(customer_id)
    if customer_obj is None:
        LOG.error(f"{customer_id}: customer not found")
        context.abort()

    for panel_id in panels:
        panel_obj = status.panel(panel_id)
        if panel_obj is None:
            LOG.error(f"{panel_id}: panel not found")
            context.abort()

    new_family = status.add_family(name=name, panels=panels, priority=priority)
    new_family.customer = customer_obj
    status.add_commit(new_family)
    LOG.info(f"{new_family.internal_id}: new family added")


@add.command()
@click.option("-m", "--mother", help="sample ID for mother of sample")
@click.option("-f", "--father", help="sample ID for father of sample")
@click.option("-s", "--status", type=click.Choice(STATUS_OPTIONS), required=True)
@click.argument("family_id")
@click.argument("sample_id")
@click.pass_context
def relationship(context, mother, father, status, family_id, sample_id):
    """Create a link between a FAMILY_ID and a SAMPLE_ID."""
    status_db = context.obj["db"]
    mother_obj = None
    father_obj = None
    family_obj = status_db.family(family_id)
    if family_obj is None:
        LOG.error("%s: family not found", family_id)
        context.abort()

    sample_obj = status_db.sample(sample_id)
    if sample_obj is None:
        LOG.error("%s: sample not found", sample_id)
        context.abort()

    if mother:
        mother_obj = status_db.sample(mother)
        if mother_obj is None:
            LOG.error("%s: mother not found", mother)
            context.abort()

    if father:
        father_obj = status_db.sample(father)
        if father_obj is None:
            LOG.error("%s: father not found", father)
            context.abort()

    new_record = status_db.relate_sample(
        family_obj, sample_obj, status, mother=mother_obj, father=father_obj
    )
    status_db.add_commit(new_record)
    LOG.info(f"related {family_obj.internal_id} to {sample_obj.internal_id}")



"""Test the upload genotype command"""

import logging

from click.testing import CliRunner

from cg.cli.upload.genotype import genotypes as upload_genotypes_cmd


def test_upload_genotype(
    upload_genotypes_context: dict, case_id: str, cli_runner: CliRunner, caplog
):
    """Test to upload genotypes via the CLI"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a case that is ready for upload sequence genotypes
    family_obj = upload_genotypes_context["status"].family(case_id)
    assert family_obj

    # WHEN uploading the genotypes
    result = cli_runner.invoke(upload_genotypes_cmd, [case_id], obj=upload_genotypes_context)
    # THEN check that the command exits with success
    assert result.exit_code == 0
    # THEN assert the correct information is communicated
    assert "loading VCF genotypes for sample(s):" in caplog.text



"""Set family attributes in the status database"""
import logging

import click
from cg.constants import FAMILY_ACTIONS, PRIORITY_OPTIONS

LOG = logging.getLogger(__name__)


@click.command()
@click.option("-a", "--action", type=click.Choice(FAMILY_ACTIONS), help="update family action")
@click.option("-c", "--customer-id", type=click.STRING, help="update customer")
@click.option("-g", "--panel", "panels", multiple=True, help="update gene panels")
@click.option("-p", "--priority", type=click.Choice(PRIORITY_OPTIONS), help="update priority")
@click.argument("family_id")
@click.pass_context
def family(
    context: click.Context,
    action: str,
    priority: str,
    panels: [str],
    family_id: str,
    customer_id: str,
):
    """Update information about a family."""

    family_obj = context.obj["status"].family(family_id)
    if family_obj is None:
        LOG.error("Can't find family %s,", family_id)
        raise click.Abort
    if not any([action, panels, priority, customer_id]):
        LOG.error("Nothing to change")
        raise click.Abort
    if action:
        LOG.info(f"Update action: {family_obj.action or 'NA'} -> {action}")
        family_obj.action = action
    if customer_id:
        customer_obj = context.obj["status"].customer(customer_id)
        if customer_obj is None:
            LOG.error("Unknown customer: %s", customer_id)
            raise click.Abort
        LOG.info(f"Update customer: {family_obj.customer.internal_id} -> {customer_id}")
        family_obj.customer = customer_obj
    if panels:
        for panel_id in panels:
            panel_obj = context.obj["status"].panel(panel_id)
            if panel_obj is None:
                LOG.error(f"unknown gene panel: {panel_id}")
                raise click.Abort
        LOG.info(f"Update panels: {', '.join(family_obj.panels)} -> {', '.join(panels)}")
        family_obj.panels = panels
    if priority:
        LOG.info(f"update priority: {family_obj.priority_human} -> {priority}")
        family_obj.priority_human = priority

    context.obj["status"].commit()



"""Utility functions to simply add test data in a cg store"""
import logging
from datetime import datetime
from typing import List

from housekeeper.store import models as hk_models

from cg.apps.hk import HousekeeperAPI
from cg.store import Store, models

LOG = logging.getLogger(__name__)


class StoreHelpers:
    """Class to hold helper functions that needs to be used all over"""

    @staticmethod
    def ensure_hk_bundle(store: HousekeeperAPI, bundle_data: dict) -> hk_models.Bundle:
        """Utility function to add a bundle of information to a housekeeper api"""
        bundle_exists = False
        for bundle in store.bundles():
            if bundle.name != bundle_data["name"]:
                continue
            bundle_exists = True
        if not bundle_exists:
            _bundle, _version = store.add_bundle(bundle_data)
            store.add_commit(_bundle, _version)
        return _bundle

    def ensure_hk_version(self, store: HousekeeperAPI, bundle_data: dict) -> hk_models.Version:
        """Utility function to return existing or create an version for tests"""
        _bundle = self.ensure_hk_bundle(store, bundle_data)
        _version = store.last_version(_bundle.name)
        return _version

    def ensure_application_version(
        self,
        store: Store,
        application_tag: str = "dummy_tag",
        application_type: str = "wgs",
        is_external: bool = False,
        is_rna: bool = False,
        description: str = None,
        sequencing_depth: int = None,
        is_accredited: bool = False,
    ) -> models.ApplicationVersion:
        """Utility function to return existing or create application version for tests"""
        if is_rna:
            application_tag = "rna_tag"
            application_type = "wts"

        application = store.application(tag=application_tag)
        if not application:
            application = self.add_application(
                store,
                application_tag,
                application_type,
                is_external=is_external,
                description=description,
                is_accredited=is_accredited,
                sequencing_depth=sequencing_depth,
            )

        prices = {"standard": 10, "priority": 20, "express": 30, "research": 5}
        version = store.application_version(application, 1)
        if not version:
            version = store.add_version(application, 1, valid_from=datetime.now(), prices=prices)

            store.add_commit(version)
        return version

    @staticmethod
    def add_application(
        store: Store,
        application_tag: str = "dummy_tag",
        application_type: str = "wgs",
        description: str = None,
        is_accredited: bool = False,
        is_external: bool = False,
        **kwargs,
    ) -> models.Application:
        """Utility function to add a application to a store"""
        application = store.application(tag=application_tag)
        if application:
            return application

        if not description:
            description = "dummy_description"
        application = store.add_application(
            tag=application_tag,
            category=application_type,
            description=description,
            percent_kth=80,
            is_accredited=is_accredited,
            limitations="A limitation",
            is_external=is_external,
            **kwargs,
        )
        store.add_commit(application)
        return application

    @staticmethod
    def ensure_bed_version(store: Store, bed_name: str = "dummy_bed") -> models.ApplicationVersion:
        """Utility function to return existing or create bed version for tests"""
        bed = store.bed(name=bed_name)
        if not bed:
            bed = store.add_bed(name=bed_name)
            store.add_commit(bed)

        version = store.latest_bed_version(bed_name)
        if not version:
            version = store.add_bed_version(bed, 1, "dummy_filename", shortname=bed_name)
            store.add_commit(version)
        return version

    @staticmethod
    def ensure_customer(
        store: Store,
        customer_id: str = "cust000",
        name: str = "Production",
        scout_access: bool = False,
        customer_group: str = "all_customers",
    ) -> models.Customer:
        """Utility function to return existing or create customer for tests"""
        customer_group_id = customer_group or customer_id + "_group"
        customer_group = store.customer_group(customer_group_id)
        if not customer_group:
            customer_group = store.add_customer_group(customer_group_id, customer_group_id)

        customer = store.customer(customer_id)

        if not customer:
            customer = store.add_customer(
                internal_id=customer_id,
                name=name,
                scout_access=scout_access,
                customer_group=customer_group,
                invoice_address="Test street",
                invoice_reference="ABCDEF",
            )
            store.add_commit(customer)
        return customer

    def add_analysis(
        self,
        store: Store,
        family: models.Family = None,
        started_at: datetime = None,
        completed_at: datetime = None,
        uploaded_at: datetime = None,
        upload_started: datetime = None,
        delivery_reported_at: datetime = None,
        cleaned_at: datetime = None,
        pipeline: str = "dummy_pipeline",
        pipeline_version: str = "1.0",
        uploading: bool = False,
        config_path: str = None,
    ) -> models.Analysis:
        """Utility function to add an analysis for tests"""

        if not family:
            family = self.add_family(store)

        analysis = store.add_analysis(pipeline=pipeline, version=pipeline_version)

        analysis.started_at = started_at or datetime.now()
        if completed_at:
            analysis.completed_at = completed_at
        if uploaded_at:
            analysis.uploaded_at = uploaded_at
        if delivery_reported_at:
            analysis.delivery_report_created_at = delivery_reported_at
        if cleaned_at:
            analysis.cleaned_at = cleaned_at
        if uploading:
            analysis.upload_started_at = upload_started or datetime.now()
        if config_path:
            analysis.config_path = config_path
        if pipeline:
            analysis.pipeline = pipeline

        analysis.limitations = "A limitation"
        analysis.family = family
        store.add_commit(analysis)
        return analysis

    def add_sample(
        self,
        store: Store,
        sample_id: str = "sample_test",
        internal_id: str = None,
        gender: str = "female",
        is_tumour: bool = False,
        is_rna: bool = False,
        is_external: bool = False,
        data_analysis: str = "balsamic",
        application_tag: str = "dummy_tag",
        application_type: str = "tgs",
        customer_name: str = None,
        reads: int = None,
        loqus_id: str = None,
        **kwargs,
    ) -> models.Sample:
        """Utility function to add a sample to use in tests"""
        customer_name = customer_name or "cust000"
        customer = self.ensure_customer(store, customer_name)
        application_version = self.ensure_application_version(
            store,
            application_tag=application_tag,
            application_type=application_type,
            is_external=is_external,
            is_rna=is_rna,
        )
        application_version_id = application_version.id
        sample = store.add_sample(
            name=sample_id,
            sex=gender,
            tumour=is_tumour,
            sequenced_at=datetime.now(),
            data_analysis=data_analysis,
            reads=reads,
        )

        sample.application_version_id = application_version_id
        sample.customer = customer
        sample.is_external = is_external

        if loqus_id:
            sample.loqusdb_id = loqus_id

        if kwargs.get("delivered_at"):
            sample.delivered_at = kwargs["delivered_at"]

        if kwargs.get("received_at"):
            sample.received_at = kwargs["received_at"]

        if kwargs.get("prepared_at"):
            sample.received_at = kwargs["prepared_at"]

        if kwargs.get("flowcell"):
            sample.flowcells.append(kwargs["flowcell"])

        if internal_id:
            sample.internal_id = internal_id

        store.add_commit(sample)
        return sample

    def ensure_panel(
        self, store: Store, panel_id: str = "panel_test", customer_id: str = "cust000"
    ) -> models.Panel:
        """Utility function to add a panel to use in tests"""
        customer = self.ensure_customer(store, customer_id)
        panel = store.panel(panel_id)
        if not panel:
            panel = store.add_panel(
                customer=customer,
                name=panel_id,
                abbrev=panel_id,
                version=1.0,
                date=datetime.now(),
                genes=1,
            )
            store.add_commit(panel)
        return panel

    def add_family(
        self,
        store: Store,
        family_id: str = "family_test",
        internal_id: str = None,
        customer_id: str = "cust000",
        panels: List = None,
        family_obj: models.Family = None,
    ) -> models.Family:
        """Utility function to add a family to use in tests,
        If no family object is used a autogenerated family id will be used

        """
        customer = self.ensure_customer(store, customer_id)
        if family_obj:
            panels = family_obj.panels
        if not panels:
            panels = ["panel_test"]
        for panel_name in panels:
            self.ensure_panel(store, panel_id=panel_name, customer_id=customer_id)

        if not family_obj:
            family_obj = store.add_family(name=family_id, panels=panels)

        if internal_id:
            family_obj.internal_id = internal_id

        family_obj.customer = customer
        store.add_commit(family_obj)
        return family_obj

    def ensure_family(self, store: Store, name: str, customer: models.Customer):
        family = store.find_family(customer=customer, name=name)
        if not family:
            family = store.add_family(name=name, panels=None)
            family.customer = customer
        return family

    def ensure_family_from_dict(
        self,
        store: Store,
        family_info: dict,
        app_tag: str = None,
        ordered_at: datetime = None,
        completed_at: datetime = None,
    ):
        """Load a family with samples and link relations"""
        customer_obj = self.ensure_customer(store)
        family_obj = store.Family(
            name=family_info["name"],
            panels=family_info["panels"],
            internal_id=family_info["internal_id"],
            priority="standard",
            ordered_at=ordered_at,
        )

        family_obj = self.add_family(
            store, family_obj=family_obj, customer_id=customer_obj.internal_id
        )

        app_tag = app_tag or "WGTPCFC030"
        app_type = family_info.get("application_type", "wgs")
        self.ensure_application_version(store, application_tag=app_tag)
        data_analysis = family_info.get("data_analysis", "mip")
        sample_objs = {}
        for sample_data in family_info["samples"]:
            sample_id = sample_data["internal_id"]
            sample_obj = self.add_sample(
                store,
                customer_name=sample_data["name"],
                gender=sample_data["sex"],
                internal_id=sample_id,
                data_analysis=data_analysis,
                application_type=app_type,
                ticket=sample_data["ticket_number"],
                reads=sample_data["reads"],
            )
            sample_objs[sample_id] = sample_obj

        for sample_data in family_info["samples"]:
            sample_obj = sample_objs[sample_data["internal_id"]]
            father = None
            if sample_data.get("father"):
                father = sample_objs[sample_data["father"]]
            mother = None
            if sample_data.get("mother"):
                mother = sample_objs[sample_data["mother"]]
            self.add_relationship(
                store,
                family=family_obj,
                sample=sample_obj,
                status=sample_data.get("status", "unknown"),
                father=father,
                mother=mother,
            )

        self.add_analysis(
            store,
            pipeline="pipeline",
            family=family_obj,
            completed_at=completed_at or datetime.now(),
        )
        return family_obj

    @staticmethod
    def add_organism(
        store: Store,
        internal_id: str = "organism_test",
        name: str = "organism_name",
        reference_genome: str = "reference_genome_test",
    ) -> models.Organism:
        """Utility function to add an organism to use in tests"""
        organism = store.add_organism(
            internal_id=internal_id, name=name, reference_genome=reference_genome
        )
        return organism

    def ensure_organism(
        self,
        store: Store,
        organism_id: str = "organism_test",
        name: str = "organism_name",
        reference_genome: str = "reference_genome_test",
    ) -> models.Organism:
        """Utility function to add an organism to use in tests"""
        organism = self.add_organism(
            store, internal_id=organism_id, name=name, reference_genome=reference_genome
        )
        store.add_commit(organism)
        return organism

    def add_microbial_sample(
        self,
        store: Store,
        sample_id: str = "microbial_sample_id",
        priority: str = "research",
        name: str = "microbial_name_test",
        organism: models.Organism = None,
        comment: str = "comment",
        ticket: int = 123456,
    ) -> models.Sample:
        """Utility function to add a sample to use in tests"""
        customer = self.ensure_customer(store, "cust_test")
        application_version = self.ensure_application_version(store)
        if not organism:
            organism = self.ensure_organism(store)

        sample = store.add_sample(
            application_version=application_version,
            comment=comment,
            internal_id=sample_id,
            name=name,
            organism=organism,
            priority=priority,
            reads=6000000,
            sex="unknown",
            ticket=ticket,
        )
        sample.customer = customer
        case = self.ensure_family(store=store, name=str(ticket), customer=customer)
        self.add_relationship(store=store, family=case, sample=sample)
        return sample

    def add_samples(self, store: Store, nr_samples: int = 5) -> list:
        """Utility function to add a number of samples to use in tests"""
        samples = []
        if nr_samples < 2:
            nr_samples = 2
        for sample_id in range(1, nr_samples):
            samples.append(self.add_sample(store, str(sample_id)))
        return samples

    @staticmethod
    def add_flowcell(
        store: Store,
        flowcell_id: str = "flowcell_test",
        archived_at: datetime = None,
        samples: list = None,
    ) -> models.Flowcell:
        """Utility function to set a flowcell to use in tests"""
        flowcell_obj = store.add_flowcell(
            name=flowcell_id,
            sequencer="dummy_sequencer",
            sequencer_type="hiseqx",
            date=datetime.now(),
        )
        flowcell_obj.archived_at = archived_at
        if samples:
            flowcell_obj.samples = samples

        store.add_commit(flowcell_obj)
        return flowcell_obj

    @staticmethod
    def add_relationship(
        store: Store,
        sample: models.Sample,
        family: models.Family,
        status: str = "unknown",
        father: models.Sample = None,
        mother: models.Sample = None,
    ) -> models.FamilySample:
        """Utility function to link a sample to a family"""
        link = store.relate_sample(
            sample=sample, family=family, status=status, father=father, mother=mother
        )
        store.add_commit(link)
        return link



"""Commands for compressing files, and cleaning uncompressed files"""
import logging

import click

from cg.apps import crunchy, hk
from cg.meta.compress import CompressAPI
from cg.store import Store

from .fastq import clean_fastq, decompress_spring, fastq_cmd, fix_spring

LOG = logging.getLogger(__name__)


@click.group()
@click.pass_context
def compress(context):
    """Compress files"""
    context.obj["db"] = Store(context.obj.get("database"))

    hk_api = hk.HousekeeperAPI(context.obj)
    crunchy_api = crunchy.CrunchyAPI(context.obj)

    compress_api = CompressAPI(hk_api=hk_api, crunchy_api=crunchy_api)
    context.obj["compress"] = compress_api


compress.add_command(fastq_cmd)


@compress.group()
def clean():
    """Clean uncompressed files"""


clean.add_command(clean_fastq)
clean.add_command(fix_spring)


@compress.group()
@click.pass_context
def decompress(context):
    """Decompress files"""
    context.obj["db"] = Store(context.obj.get("database"))

    hk_api = hk.HousekeeperAPI(context.obj)
    crunchy_api = crunchy.CrunchyAPI(context.obj)

    compress_api = CompressAPI(hk_api=hk_api, crunchy_api=crunchy_api)
    context.obj["compress"] = compress_api


decompress.add_command(decompress_spring)



"""Click commands to store mip analyses"""
import logging
from pathlib import Path

import click

from cg.apps import hk, tb
from cg.exc import (
    AnalysisNotFinishedError,
    AnalysisDuplicationError,
    BundleAlreadyAddedError,
    PipelineUnknownError,
    MandatoryFilesMissing,
)
from cg.meta.store.base import gather_files_and_bundle_in_housekeeper
from cg.store import Store
from cg.constants import EXIT_SUCCESS, EXIT_FAIL


LOG = logging.getLogger(__name__)


@click.group()
@click.pass_context
def store(context):
    """Store results from MIP in housekeeper."""
    context.obj["db"] = Store(context.obj["database"])
    context.obj["tb_api"] = tb.TrailblazerAPI(context.obj)
    context.obj["hk_api"] = hk.HousekeeperAPI(context.obj)


@store.command()
@click.argument("config-stream", type=click.File("r"), required=False)
@click.pass_context
def analysis(context, config_stream):
    """Store a finished analysis in Housekeeper."""
    status = context.obj["db"]
    hk_api = context.obj["hk_api"]

    exit_code = EXIT_SUCCESS
    if not config_stream:
        LOG.error("Provide a config file.")
        raise click.Abort

    try:
        new_analysis = gather_files_and_bundle_in_housekeeper(
            config_stream,
            hk_api,
            status,
            workflow="mip",
        )
    except (
        AnalysisNotFinishedError,
        AnalysisDuplicationError,
        BundleAlreadyAddedError,
        PipelineUnknownError,
        MandatoryFilesMissing,
    ) as error:
        click.echo(click.style(error.message, fg="red"))
        exit_code = EXIT_FAIL
    except FileNotFoundError as error:
        click.echo(click.style(f"missing file: {error.args[0]}", fg="red"))
        exit_code = EXIT_FAIL
    if exit_code:
        raise click.Abort
    status.add_commit(new_analysis)
    click.echo(click.style("included files in Housekeeper", fg="green"))


@store.command()
@click.pass_context
def completed(context):
    """Store all completed analyses."""
    hk_api = context.obj["hk_api"]
    tb_api = context.obj["tb_api"]

    exit_code = EXIT_SUCCESS
    for analysis_obj in tb_api.analyses(status="completed", deleted=False):
        existing_record = hk_api.version(analysis_obj.family, analysis_obj.started_at)
        if existing_record:
            LOG.debug("analysis stored: %s - %s", analysis_obj.family, analysis_obj.started_at)
            continue
        click.echo(click.style(f"storing family: {analysis_obj.family}", fg="blue"))
        with Path(analysis_obj.config_path).open() as config_stream:
            try:
                context.invoke(analysis, config_stream=config_stream)
            except (Exception, click.Abort):
                LOG.error("case storage failed: %s", analysis_obj.family, exc_info=True)
                exit_code = EXIT_FAIL
    if exit_code:
        raise click.Abort



"""Fixtures for cli analysis tests"""

import pytest

from cg.store import Store, models


@pytest.fixture
def base_context(analysis_store) -> dict:
    """context to use in cli"""
    return {"db": analysis_store}


@pytest.fixture(name="workflow_case_id")
def fixture_workflow_case_id() -> str:
    """Return a special case id"""
    return "dna_case"


@pytest.fixture(name="dna_sample_id")
def fixture_dna_sample_id() -> str:
    """Return a special sample id"""
    return "dna_sample"


@pytest.fixture(name="rna_sample_id")
def fixture_rna_sample_id() -> str:
    """Return a special sample id"""
    return "rna_sample"


@pytest.fixture(scope="function", name="analysis_store")
def fixture_analysis_store(base_store: Store, workflow_case_id, helpers) -> Store:
    """Store to be used in tests"""
    _store = base_store

    case = helpers.add_family(_store, workflow_case_id)

    sample = helpers.add_sample(_store, "dna_sample", is_rna=False, data_analysis="mip")
    helpers.add_relationship(_store, sample=sample, family=case)

    case = helpers.add_family(_store, "rna_case")
    sample = helpers.add_sample(_store, "rna_sample", is_rna=True, data_analysis="mip")
    helpers.add_relationship(_store, sample=sample, family=case)

    return _store


@pytest.fixture(scope="function")
def dna_case(analysis_store, helpers) -> models.Family:
    """Case with DNA application"""
    cust = helpers.ensure_customer(analysis_store)
    return analysis_store.find_family(cust, "dna_case")


@pytest.fixture(scope="function")
def rna_case(analysis_store, helpers) -> models.Family:
    """Case with RNA application"""
    cust = helpers.ensure_customer(analysis_store)
    return analysis_store.find_family(cust, "rna_case")


class MockTB:
    """Trailblazer mock fixture"""

    def __init__(self):
        self._link_was_called = False
        self._mark_analyses_deleted_called = False
        self._add_pending_was_called = False
        self._add_pending_analysis_was_called = False
        self._family = None
        self._temp = None
        self._case_id = None
        self._email = None
        self._status = None

    def analyses(self, family=None, status=None, temp=None):
        """Mock TB analyses models"""

        self._family = family
        self._status = status
        self._temp = temp

        class Row:
            """Mock a record representing an analysis"""

            def __init__(self):
                """We need to initialize _first_was_called
                so that we can set it in `first()` and retrieve
                it in `first_was_called()`. This way we can easily
                run the invoking code and make sure the function was
                called.
                """

                self._first_was_called = False

            def first(self):
                """Mock that the first row doesn't exist"""
                self._first_was_called = True

            def first_was_called(self):
                """Check if first was called"""
                return self._first_was_called

        return Row()

    def mark_analyses_deleted(self, case_id: str):
        """Mock this function"""
        self._case_id = case_id
        self._mark_analyses_deleted_called = True

    def add_pending(self, case_id: str, email: str):
        """Mock this function"""
        self._case_id = case_id
        self._email = email
        self._add_pending_was_called = True

    def add_pending_analysis(self, case_id: str, email: str):
        """Mock adding a pending analyses"""
        self._case_id = case_id
        self._email = email
        self._add_pending_analysis_was_called = True

    def mark_analyses_deleted_called(self):
        """check if mark_analyses_deleted was called"""
        return self._mark_analyses_deleted_called

    def add_pending_was_called(self):
        """check if add_pending was called"""
        return self._add_pending_was_called

    def is_analysis_ongoing(self, case_id: str):
        """Override TrailblazerAPI is_ongoing method to avoid default behaviour"""
        return False

    def is_analysis_failed(self, case_id: str):
        """Override TrailblazerAPI is_failed method to avoid default behaviour"""
        return False

    def is_analysis_completed(self, case_id: str):
        """Override TrailblazerAPI is_completed method to avoid default behaviour"""
        return False

    def get_analysis_status(self, case_id: str):
        """Override TrailblazerAPI get_analysis_status method to avoid default behaviour"""
        return None

    def has_analysis_started(self, case_id: str):
        """Override TrailblazerAPI has_analysis_started method to avoid default behaviour"""
        return False


@pytest.fixture(scope="function")
def tb_api():
    """Trailblazer API fixture"""

    return MockTB()



"""Tests the Presenter class"""
from cg.meta.report.presenter import Presenter


def test_present_float_string():
    # GIVEN

    # WHEN formatting None
    presentable_string = Presenter().process_float_string(None, 0)

    # THEN we should get 'N/A' back
    assert presentable_string == Presenter().DEFAULT_NA


def test_present_date():
    # GIVEN

    # WHEN formatting None
    presentable_string = Presenter().process_datetime(None)

    # THEN we should get 'N/A' back
    assert presentable_string == Presenter().DEFAULT_NA


def test_present_string():
    # GIVEN

    # WHEN formatting None
    presentable_string = Presenter().process_string(None)

    # THEN we should get 'N/A' back
    assert presentable_string == Presenter().DEFAULT_NA


def test_present_int():
    # GIVEN

    # WHEN formatting None
    presentable_string = Presenter().process_int(None)

    # THEN we should get 'N/A' back
    assert presentable_string == Presenter().DEFAULT_NA


def test_present_set():
    # GIVEN

    # WHEN formatting None
    presentable_string = Presenter().process_set(None)

    # THEN we should get 'N/A' back
    assert presentable_string == Presenter().DEFAULT_NA


def test_present_dict():
    # GIVEN a dict to present
    dict_to_present = {
        "nested_dict": {"string_key": "string_value", "none_key": None, "empty_key": ""}
    }

    # WHEN formatting dict
    presentable_dict = Presenter().process_dict(dict_to_present)

    # THEN we should get 'N/A' back
    assert presentable_dict["nested_dict"]["string_key"] == "string_value"
    assert presentable_dict["nested_dict"]["none_key"] == Presenter().DEFAULT_NA
    assert presentable_dict["nested_dict"]["empty_key"] == Presenter().DEFAULT_NA



"""This script tests the cli methods to add families to status-db"""
from datetime import datetime

from cg.store import Store


def test_add_family_required(invoke_cli, disk_store: Store):
    """Test to add a family using only the required arguments"""
    # GIVEN a database with a customer and an panel

    # WHEN adding a family
    db_uri = disk_store.uri
    customer_id = add_customer(disk_store)
    panel_id = add_panel(disk_store)
    name = "family_name"

    result = invoke_cli(
        ["--database", db_uri, "add", "family", "--panel", panel_id, customer_id, name]
    )

    # THEN then it should be added
    assert result.exit_code == 0
    assert disk_store.Family.query.count() == 1
    assert disk_store.Family.query.first().name == name
    assert disk_store.Family.query.first().panels == [panel_id]


def test_add_family_bad_customer(invoke_cli, disk_store: Store):
    """Test to add a family using a non-existing customer"""
    # GIVEN an empty database

    # WHEN adding a family
    db_uri = disk_store.uri
    panel_id = "dummy_panel"
    customer_id = "dummy_customer"
    name = "dummy_name"
    result = invoke_cli(
        ["--database", db_uri, "add", "family", "--panel", panel_id, customer_id, name]
    )

    # THEN then it should complain about missing customer instead of adding a family
    assert result.exit_code == 1
    assert disk_store.Family.query.count() == 0


def test_add_family_bad_panel(invoke_cli, disk_store: Store):
    """Test to add a family using a non-existing panel"""
    # GIVEN a database with a customer

    # WHEN adding a family
    db_uri = disk_store.uri
    panel_id = "dummy_panel"
    customer_id = add_customer(disk_store)
    name = "dummy_name"
    result = invoke_cli(
        ["--database", db_uri, "add", "family", "--panel", panel_id, customer_id, name]
    )

    # THEN then it should complain about missing panel instead of adding a family
    assert result.exit_code == 1
    assert disk_store.Family.query.count() == 0


def test_add_family_priority(invoke_cli, disk_store: Store):
    """Test that the added family get the priority we send in"""
    # GIVEN a database with a customer and an panel

    # WHEN adding a family
    db_uri = disk_store.uri

    customer_id = add_customer(disk_store)
    panel_id = add_panel(disk_store)
    name = "family_name"
    priority = "priority"

    result = invoke_cli(
        [
            "--database",
            db_uri,
            "add",
            "family",
            "--panel",
            panel_id,
            "--priority",
            priority,
            customer_id,
            name,
        ]
    )

    # THEN then it should be added
    assert result.exit_code == 0
    assert disk_store.Family.query.count() == 1
    assert disk_store.Family.query.first().priority_human == priority


def add_customer(disk_store, customer_id="cust_test"):
    customer_group = disk_store.add_customer_group("dummy_group", "dummy group")
    customer = disk_store.add_customer(
        internal_id=customer_id,
        name="Test Customer",
        scout_access=False,
        customer_group=customer_group,
        invoice_address="dummy_address",
        invoice_reference="dummy_reference",
    )
    disk_store.add_commit(customer)
    return customer_id


def add_panel(disk_store, panel_id="panel_test", customer_id="cust_test"):
    """utility function to add a panel to use in tests"""
    customer = disk_store.customer(customer_id)
    panel = disk_store.add_panel(
        customer=customer,
        name=panel_id,
        abbrev=panel_id,
        version=1.0,
        date=datetime.now(),
        genes=1,
    )
    disk_store.add_commit(panel)
    return panel_id


def add_application(disk_store, application_tag="dummy_tag"):
    """utility function to add an application to use in tests"""
    application = disk_store.add_application(
        tag=application_tag,
        category="wgs",
        description="dummy_description",
        percent_kth=80,
    )
    disk_store.add_commit(application)
    prices = {"standard": 10, "priority": 20, "express": 30, "research": 5}
    version = disk_store.add_version(application, 1, valid_from=datetime.now(), prices=prices)

    disk_store.add_commit(version)
    return application_tag



"""Tests for helper functions in cg compress cli"""

import os
from pathlib import Path

from cg.cli.compress import helpers


def test_get_true_dir_no_symlinks(project_dir):
    """Test to get a true dir when there are no symlinks"""
    # GIVEN a directory with some files but no symlinked files
    a_file = project_dir / "hello.txt"
    a_file.touch()
    assert a_file.exists()
    # WHEN fetching the true dir for the files in fixture dir
    true_dir = helpers.get_true_dir(a_file.parent)
    # THEN assert that the true_dir is None since there where no symbolic links in the project_dir
    assert true_dir is None


def test_get_true_dir_when_symlinks(project_dir):
    """Test to get a true dir when there are symlinks to another directory

    The function should return the path to another directory when there are symlinks in one
    """
    # GIVEN a directory with a file with some content
    content = "hello world"
    a_file = project_dir / "hello.txt"
    a_file.write_text(content)
    assert a_file.exists()
    # GIVEN a subdirectory with a link to the above file
    new_dir = project_dir / "new_dir/"
    new_dir.mkdir()
    a_link = new_dir / "link_to_hello"
    a_link.symlink_to(a_file)
    assert a_link.read_text() == content
    assert a_link.parent == new_dir
    assert a_link.is_symlink() is True

    # WHEN fetching the true dir for the symlinked dir
    true_dir = helpers.get_true_dir(a_link.parent)

    # THEN assert that the true_dir the same as the parent of the destination of the symlink
    assert true_dir == project_dir


def test_get_versions_no_bundle(housekeeper_api):
    """Test to get latest versions of bundles when no bundles exists"""
    # GIVEN a empty housekeeper_api

    # WHEN fetching versions
    versions = helpers.get_versions(housekeeper_api)

    # THEN assert no versions was returned
    assert sum(1 for version in versions) == 0


def test_get_versions_one_bundle(housekeeper_api, spring_bundle):
    """Test to get latest versions of bundles when no bundles exists"""
    # GIVEN a populated housekeeper_api
    housekeeper_api.add_bundle(spring_bundle)

    # WHEN fetching versions
    versions = helpers.get_versions(housekeeper_api)

    # THEN assert no versions was returned
    assert sum(1 for version in versions) == 1


def test_correct_spring_paths(
    housekeeper_api, spring_bundle_symlink_problem, symlinked_fastqs, new_dir
):
    """docstring for test_correct_spring_paths"""
    # GIVEN a populated housekeeper_api
    housekeeper_api.add_bundle(spring_bundle_symlink_problem)
    # GIVEN that the spring files exists in the wrong location
    versions = helpers.get_versions(housekeeper_api)
    version = next(versions)
    for file_path in version.files:
        assert not Path(file_path.full_path).exists()

    # WHEN updating the spring paths
    helpers.correct_spring_paths(housekeeper_api)

    # THEN assert that the spring paths has been moved
    for file_path in version.files:
        assert Path(file_path.full_path).exists()



""" Microsalt specific functionality for storing files in Housekeeper """
import datetime as dt
import logging
from pathlib import Path
import ruamel.yaml

from _io import TextIOWrapper

from cg.constants import HK_TAGS, MICROSALT_TAGS
from cg.meta.store import base as store_base

LOG = logging.getLogger(__name__)


def add_microbial_analysis(config_stream: TextIOWrapper) -> dict:
    """Gather information from microSALT analysis to store."""

    deliverables = ruamel.yaml.safe_load(config_stream)
    new_bundle = build_microbial_bundle(deliverables)

    return new_bundle


def build_microbial_bundle(deliverables: dict) -> dict:
    """Create a new bundle to store in Housekeeper"""

    project_name = _get_microbial_name(deliverables)
    files = store_base.deliverables_files(deliverables, analysis_type="microbial")

    data = {
        "name": str(project_name),
        "created": _get_date_from_results_path(deliverables),
        "files": files,
    }
    return data


def microbial_deliverables_files(deliverables: dict) -> list:
    """Get all deliverable files from the pipeline"""

    pipeline_tags = HK_TAGS["microbial"]
    analysis_type_tags = MICROSALT_TAGS
    files = store_base.parse_files(deliverables, pipeline_tags, analysis_type_tags)

    return files


def _get_microbial_name(deliverables: dict) -> str:
    """ Get microbial id from deliverables """
    for file_ in deliverables["files"]:
        if file_["tag"] == "sampleinfo":
            return file_["id"]
    return None


def _get_date_from_results_path(deliverables: dict) -> dt.datetime:
    """ Get date from results path """
    first_file, *_ = deliverables["files"]
    results_dir = Path(first_file["path"]).parent.name
    [_, date, time] = results_dir.split("_")
    [year, month, day] = list(map(int, date.split(".")))
    [hour, minutes, seconds] = list(map(int, time.split(".")))
    return dt.datetime(year, month, day, hour, minutes, seconds)



"""This script tests the cli mip store functions"""
import logging

from click.testing import CliRunner

from cg.cli.workflow.mip.store import analysis, completed
from cg.constants import EXIT_FAIL, EXIT_SUCCESS


def test_store_no_config(cli_runner: CliRunner, mip_store_context: dict, caplog):
    """Test the function to check for config files"""

    caplog.set_level(logging.INFO)
    # WHEN we run store analysis on no config
    result = cli_runner.invoke(analysis, obj=mip_store_context)
    # THEN We should be informed to enter a config file
    assert "Provide a config file." in caplog.text
    # THEN the exit code should be EXIT_FAIL
    assert result.exit_code == EXIT_FAIL


def test_store_analysis_exception(
    cli_runner: CliRunner, mip_store_context: dict, mip_configs: dict
):
    """Test that the analysis function enters the exception clause"""

    # WHEN we run store on a case without deliverables
    result = cli_runner.invoke(analysis, [str(mip_configs["purplesnail"])], obj=mip_store_context)
    # THEN we should be informed that mandatory files are missing
    assert "Mandatory files are missing" in result.output
    # THEN we should not include any files in housekeeper
    assert "included files in Housekeeper" not in result.output
    # THEN the exit code should be EXIT_FAIL
    assert result.exit_code == EXIT_FAIL


def test_store_analysis(
    cli_runner: CliRunner,
    mip_configs: dict,
    mip_store_context: dict,
    mip_qc_sample_info: dict,
    caplog,
):
    """Test if store completed stores a completed sample"""

    caplog.set_level(logging.INFO)
    # WHEN we run store on a config
    result = cli_runner.invoke(analysis, [str(mip_configs["yellowhog"])], obj=mip_store_context)
    # THEN we should store the files in housekeeper
    assert "new bundle added: yellowhog" in caplog.text
    assert "included files in Housekeeper" in result.output
    # THEN the exit code should be EXIT_SUCCESS
    assert result.exit_code == EXIT_SUCCESS


def test_store_completed_good_cases(
    cli_runner: CliRunner, mip_store_context: dict, mip_case_ids: dict, caplog
):
    """Test if store completed stores function"""

    caplog.set_level(logging.INFO)
    # WHEN we run store all completed cases
    result = cli_runner.invoke(completed, obj=mip_store_context)
    # THEN command is run successfully
    for case in mip_case_ids:
        assert f"storing family: {case}" in result.output
    # THEN some cases should be added and some should fail
    assert "new bundle added: yellowhog" in caplog.text
    assert "case storage failed: purplesnail" in caplog.text
    assert "new bundle added: bluezebra" in caplog.text
    assert "included files in Housekeeper" in result.output
    # THEN the command should have an EXIT_FAIL code
    assert result.exit_code == EXIT_FAIL



""" MIP specific functionality for storing files in Housekeeper """
import logging
from pathlib import Path
import ruamel.yaml

from cg.exc import AnalysisNotFinishedError
from cg.meta.store import base as store_base

LOG = logging.getLogger(__name__)


def add_mip_analysis(config_stream):
    """Gather information from MIP analysis to store."""
    config_raw = ruamel.yaml.safe_load(config_stream)
    config_data = parse_config(config_raw)
    sampleinfo_raw = ruamel.yaml.safe_load(Path(config_data["sampleinfo_path"]).open())
    sampleinfo_data = parse_sampleinfo(sampleinfo_raw)

    if sampleinfo_data["is_finished"] is False:
        raise AnalysisNotFinishedError("analysis not finished")

    deliverables_raw = ruamel.yaml.safe_load(Path(config_raw["store_file"]).open())
    new_bundle = store_base.build_bundle(config_data, sampleinfo_data, deliverables_raw)

    return new_bundle


def parse_config(data: dict) -> dict:
    """Parse MIP config file.

    Args:
        data (dict): raw YAML input from MIP analysis config file

    Returns:
        dict: parsed data
    """
    return {
        "email": data.get("email"),
        "case": data["case_id"],
        "samples": _get_sample_analysis_type(data),
        "is_dryrun": bool("dry_run_all" in data),
        "out_dir": data["outdata_dir"],
        "priority": data["slurm_quality_of_service"],
        "sampleinfo_path": data["sample_info_file"],
    }


def parse_sampleinfo(data: dict) -> dict:
    """Parse MIP sample info file.

    Args:
        data (dict): raw YAML input from MIP qc sample info file

    Returns:
        dict: parsed data
    """

    sampleinfo_data = {
        "date": data["analysis_date"],
        "is_finished": data["analysisrunstatus"] == "finished",
        "case": data["case"],
        "version": data["mip_version"],
    }

    return sampleinfo_data


def _get_sample_analysis_type(data: dict) -> list:
    """
    Get analysis type for all samples in the MIP config file
    """
    return [
        {"id": sample_id, "type": analysis_type}
        for sample_id, analysis_type in data["analysis_type"].items()
    ]



from datetime import datetime, timedelta
from typing import List

from cg.constants import PRIORITY_MAP
from cg.store import models
from cg.store.api.base import BaseHandler
from cg.utils.date import get_date
from sqlalchemy import or_, and_, func
from sqlalchemy.orm import Query

HASTA_IN_PRODUCTION = get_date("2017-09-27")


class StatusHandler(BaseHandler):
    """Handles status states for entities in the database"""

    def samples_to_recieve(self, external=False):
        """Fetch incoming samples."""
        records = (
            self.Sample.query.join(
                models.Sample.application_version, models.ApplicationVersion.application
            )
            .filter(
                models.Sample.received_at == None,
                models.Sample.downsampled_to == None,
                models.Application.is_external == external,
            )
            .order_by(models.Sample.ordered_at)
        )
        return records

    def samples_to_prepare(self):
        """Fetch samples in lab prep queue."""
        records = (
            self.Sample.query.join(
                models.Sample.application_version, models.ApplicationVersion.application
            )
            .filter(
                models.Sample.received_at != None,
                models.Sample.prepared_at == None,
                models.Sample.downsampled_to == None,
                models.Application.is_external == False,
                models.Sample.sequenced_at == None,
            )
            .order_by(models.Sample.priority.desc(), models.Sample.received_at)
        )
        return records

    def samples_to_sequence(self):
        """Fetch samples in sequencing."""
        records = (
            self.Sample.query.join(
                models.Sample.application_version, models.ApplicationVersion.application
            )
            .filter(
                models.Sample.prepared_at != None,
                models.Sample.sequenced_at == None,
                models.Sample.downsampled_to == None,
                models.Application.is_external == False,
            )
            .order_by(models.Sample.priority.desc(), models.Sample.received_at)
        )
        return records

    def cases_to_analyze(
        self, pipeline: str = "", threshold: float = None, limit: int = None
    ) -> list:
        """Returns a list if cases ready to be analyzed or set to be reanalyzed"""
        families_query = (
            self.Family.query.outerjoin(models.Analysis)
            .join(models.Family.links, models.FamilySample.sample)
            .filter(or_(models.Sample.is_external, models.Sample.sequenced_at.isnot(None)))
            .filter(models.Sample.data_analysis.ilike(f"%{pipeline}%"))
            .filter(
                or_(
                    models.Family.action == "analyze",
                    and_(models.Family.action.is_(None), models.Analysis.created_at.is_(None)),
                )
            )
            .order_by(models.Family.priority.desc(), models.Family.ordered_at)
        )

        families = [
            record
            for record in families_query
            if self._all_samples_have_sequence_data(record.links)
        ]

        if threshold:
            families = [
                record
                for record in families
                if self.all_samples_have_enough_reads(record.links, threshold=threshold)
            ]
        return families[:limit]

    def cases_to_store(self, pipeline: str, limit: int = None) -> list:
        """Returns a list of cases that may be available to store in Housekeeper"""
        families_query = (
            self.Family.query.outerjoin(models.Analysis)
            .join(models.Family.links, models.FamilySample.sample)
            .filter(or_(models.Sample.is_external, models.Sample.sequenced_at.isnot(None)))
            .filter(models.Sample.data_analysis.ilike(f"%{pipeline}%"))
            .filter(models.Family.action == "running")
            .order_by(models.Family.priority.desc(), models.Family.ordered_at)
        )
        return list(families_query)[:limit]

    def cases(
        self,
        progress_tracker=None,
        internal_id=None,
        name=None,
        days=0,
        case_action=None,
        progress_status=None,
        priority=None,
        customer_id=None,
        exclude_customer_id=None,
        data_analysis=None,
        sample_id=None,
        only_received=False,
        only_prepared=False,
        only_sequenced=False,
        only_analysed=False,
        only_uploaded=False,
        only_delivered=False,
        only_delivery_reported=False,
        only_invoiced=False,
        exclude_received=False,
        exclude_prepared=False,
        exclude_sequenced=False,
        exclude_analysed=False,
        exclude_uploaded=False,
        exclude_delivered=False,
        exclude_delivery_reported=False,
        exclude_invoiced=False,
    ):
        """Fetch cases with and w/o analyses"""
        families_q = self.Family.query

        # family filters
        if days != 0:
            filter_date = datetime.now() - timedelta(days=days)
            families_q = families_q.filter(models.Family.ordered_at > filter_date)

        if case_action:
            families_q = families_q.filter(models.Family.action == case_action)

        if priority:
            priority_db = PRIORITY_MAP[priority]
            families_q = families_q.filter(models.Family.priority == priority_db)

        if internal_id:
            families_q = families_q.filter(models.Family.internal_id.like("%" + internal_id + "%"))

        if name:
            families_q = families_q.filter(models.Family.name.like("%" + name + "%"))

        # customer filters
        if customer_id or exclude_customer_id:
            families_q = families_q.join(models.Family.customer)

            if customer_id:
                families_q = families_q.filter(models.Customer.internal_id == customer_id)

            if exclude_customer_id:
                families_q = families_q.filter(models.Customer.internal_id != exclude_customer_id)

        # sample filters
        if data_analysis or sample_id:
            families_q = families_q.join(models.Family.links, models.FamilySample.sample)
            if data_analysis:
                families_q = families_q.filter(
                    models.Sample.data_analysis.like("%" + data_analysis + "%")
                )
            if sample_id:
                families_q = families_q.filter(models.Sample.internal_id.like(sample_id))

        else:
            families_q = families_q.outerjoin(models.Family.links, models.FamilySample.sample)

        # other joins
        families_q = families_q.outerjoin(
            models.Family.analyses, models.Sample.invoice, models.Sample.flowcells
        )

        cases = []

        for record in families_q:

            samples_received = None
            samples_prepared = None
            samples_sequenced = None
            samples_delivered = None
            samples_invoiced = None
            samples_received_at = None
            samples_prepared_at = None
            samples_sequenced_at = None
            samples_delivered_at = None
            samples_invoiced_at = None
            samples_to_receive = None
            samples_to_prepare = None
            samples_to_sequence = None
            samples_to_deliver = None
            samples_to_invoice = None
            samples_received_bool = None
            samples_prepared_bool = None
            samples_sequenced_bool = None
            samples_invoiced_bool = None
            analysis_completed_at = None
            analysis_uploaded_at = None
            analysis_delivery_reported_at = None
            analysis_pipeline = None
            analysis_status = None
            analysis_completion = None
            analysis_completed_bool = None
            analysis_uploaded_bool = None
            samples_delivered_bool = None
            analysis_delivery_reported_bool = None
            samples_data_analyses = None
            flowcells_status = None
            flowcells_on_disk = None
            flowcells_on_disk_bool = None
            tat = None
            is_rerun = False

            analysis_in_progress = record.action is not None
            case_action = record.action

            total_samples = len(record.links)
            total_external_samples = len(
                [
                    link.sample.application_version.application.is_external
                    for link in record.links
                    if link.sample.application_version.application.is_external
                ]
            )
            total_internal_samples = total_samples - total_external_samples
            case_external_bool = total_external_samples == total_samples

            if total_samples > 0:
                samples_received = len(
                    [
                        link.sample.received_at
                        for link in record.links
                        if link.sample.received_at is not None
                    ]
                )
                samples_prepared = len(
                    [
                        link.sample.prepared_at
                        for link in record.links
                        if link.sample.prepared_at is not None
                    ]
                )
                samples_sequenced = len(
                    [
                        link.sample.sequenced_at
                        for link in record.links
                        if link.sample.sequenced_at is not None
                    ]
                )
                samples_delivered = len(
                    [
                        link.sample.delivered_at
                        for link in record.links
                        if link.sample.delivered_at is not None
                    ]
                )
                samples_invoiced = len(
                    [
                        link.sample.invoice.invoiced_at
                        for link in record.links
                        if link.sample.invoice and link.sample.invoice.invoiced_at
                    ]
                )

                samples_to_receive = total_internal_samples
                samples_to_prepare = total_internal_samples
                samples_to_sequence = total_internal_samples
                samples_to_deliver = total_internal_samples
                samples_to_invoice = total_samples - len(
                    [link.sample.no_invoice for link in record.links if link.sample.no_invoice]
                )

                samples_received_bool = samples_received == samples_to_receive
                samples_prepared_bool = samples_prepared == samples_to_prepare
                samples_sequenced_bool = samples_sequenced == samples_to_sequence
                samples_delivered_bool = samples_delivered == samples_to_deliver
                samples_invoiced_bool = samples_invoiced == samples_to_invoice
                samples_data_analyses = list(
                    set(link.sample.data_analysis for link in record.links)
                )

                if samples_to_receive > 0 and samples_received_bool:
                    samples_received_at = max(
                        [
                            link.sample.received_at
                            for link in record.links
                            if link.sample.received_at is not None
                        ]
                    )

                if samples_to_prepare > 0 and samples_prepared_bool:
                    samples_prepared_at = max(
                        [
                            link.sample.prepared_at
                            for link in record.links
                            if link.sample.prepared_at is not None
                        ]
                    )

                if samples_to_sequence > 0 and samples_sequenced_bool:
                    samples_sequenced_at = max(
                        [
                            link.sample.sequenced_at
                            for link in record.links
                            if link.sample.sequenced_at is not None
                        ]
                    )

                if samples_to_deliver > 0 and samples_delivered_bool:
                    samples_delivered_at = max(
                        [
                            link.sample.delivered_at
                            for link in record.links
                            if link.sample.delivered_at is not None
                        ]
                    )

                if samples_to_invoice > 0 and samples_invoiced_bool:
                    samples_invoiced_at = max(
                        [
                            link.sample.invoice.invoiced_at
                            for link in record.links
                            if link.sample.invoice and link.sample.invoice.invoiced_at
                        ]
                    )

                flowcells = len(
                    [flowcell.status for link in record.links for flowcell in link.sample.flowcells]
                )

                flowcells_status = list(
                    set(
                        flowcell.status
                        for link in record.links
                        for flowcell in link.sample.flowcells
                    )
                )
                if flowcells < total_samples:
                    flowcells_status.append("new")

                flowcells_status = ", ".join(flowcells_status)

                flowcells_on_disk = len(
                    [
                        flowcell.status
                        for link in record.links
                        for flowcell in link.sample.flowcells
                        if flowcell.status == "ondisk"
                    ]
                )

                flowcells_on_disk_bool = flowcells_on_disk == total_samples

            if record.analyses and not analysis_in_progress:
                analysis_completed_at = record.analyses[0].completed_at
                analysis_uploaded_at = record.analyses[0].uploaded_at
                analysis_delivery_reported_at = record.analyses[0].delivery_report_created_at
                analysis_pipeline = record.analyses[0].pipeline
                analysis_completed_bool = analysis_completed_at is not None
                analysis_uploaded_bool = analysis_uploaded_at is not None
                analysis_delivery_reported_bool = analysis_delivery_reported_at is not None
            elif total_samples > 0:
                analysis_completed_bool = False
                analysis_uploaded_bool = False
                analysis_delivery_reported_bool = False

            if only_received and not samples_received_bool:
                continue

            if only_prepared and not samples_prepared_bool:
                continue

            if only_sequenced and not samples_sequenced_bool:
                continue

            if only_analysed and not analysis_completed_bool:
                continue

            if only_uploaded and not analysis_uploaded_bool:
                continue

            if only_delivered and not samples_delivered_bool:
                continue

            if only_delivery_reported and not analysis_delivery_reported_bool:
                continue

            if only_invoiced and not samples_invoiced_bool:
                continue

            if exclude_received and samples_received_bool:
                continue

            if exclude_prepared and samples_prepared_bool:
                continue

            if exclude_sequenced and samples_sequenced_bool:
                continue

            if exclude_analysed and analysis_completed_bool:
                continue

            if exclude_uploaded and analysis_uploaded_bool:
                continue

            if exclude_delivered and samples_delivered_bool:
                continue

            if exclude_delivery_reported and analysis_delivery_reported_bool:
                continue

            if exclude_invoiced and samples_invoiced_bool:
                continue

            if progress_tracker:
                for analysis_obj in progress_tracker.get_latest_logged_analysis(
                    case_id=record.internal_id
                ):

                    if not analysis_status:
                        analysis_completion = round(analysis_obj.progress * 100)
                        analysis_status = analysis_obj.status

            # filter on a status
            if progress_status and progress_status != analysis_status:
                continue

            is_rerun = self._is_rerun(
                record, samples_received_at, samples_prepared_at, samples_sequenced_at
            )

            tat = self._calculate_estimated_turnaround_time(
                is_rerun,
                case_external_bool,
                record.ordered_at,
                samples_received_at,
                samples_prepared_at,
                samples_sequenced_at,
                analysis_completed_at,
                analysis_uploaded_at,
                samples_delivered_at,
            )

            max_tat = self._get_max_tat(links=record.links)

            case = {
                "internal_id": record.internal_id,
                "name": record.name,
                "ordered_at": record.ordered_at,
                "total_samples": total_samples,
                "total_external_samples": total_external_samples,
                "total_internal_samples": total_internal_samples,
                "case_external_bool": case_external_bool,
                "samples_to_receive": samples_to_receive,
                "samples_to_prepare": samples_to_prepare,
                "samples_to_sequence": samples_to_sequence,
                "samples_to_deliver": samples_to_deliver,
                "samples_to_invoice": samples_to_invoice,
                "samples_data_analyses": samples_data_analyses,
                "samples_received": samples_received,
                "samples_prepared": samples_prepared,
                "samples_sequenced": samples_sequenced,
                "samples_received_at": samples_received_at,
                "samples_prepared_at": samples_prepared_at,
                "samples_sequenced_at": samples_sequenced_at,
                "samples_delivered_at": samples_delivered_at,
                "samples_invoiced_at": samples_invoiced_at,
                "case_action": case_action,
                "analysis_status": analysis_status,
                "analysis_completion": analysis_completion,
                "analysis_completed_at": analysis_completed_at,
                "analysis_uploaded_at": analysis_uploaded_at,
                "samples_delivered": samples_delivered,
                "analysis_delivery_reported_at": analysis_delivery_reported_at,
                "samples_invoiced": samples_invoiced,
                "analysis_pipeline": analysis_pipeline,
                "samples_received_bool": samples_received_bool,
                "samples_prepared_bool": samples_prepared_bool,
                "samples_sequenced_bool": samples_sequenced_bool,
                "analysis_completed_bool": analysis_completed_bool,
                "analysis_uploaded_bool": analysis_uploaded_bool,
                "samples_delivered_bool": samples_delivered_bool,
                "analysis_delivery_reported_bool": analysis_delivery_reported_bool,
                "samples_invoiced_bool": samples_invoiced_bool,
                "flowcells_status": flowcells_status,
                "flowcells_on_disk": flowcells_on_disk,
                "flowcells_on_disk_bool": flowcells_on_disk_bool,
                "tat": tat,
                "is_rerun": is_rerun,
                "max_tat": max_tat,
            }

            cases.append(case)

        cases_sorted = sorted(cases, key=lambda k: k["tat"], reverse=True)

        return cases_sorted

    @staticmethod
    def _is_rerun(record, samples_received_at, samples_prepared_at, samples_sequenced_at):

        return (
            (len(record.analyses) > 0)
            or (samples_received_at and samples_received_at < record.ordered_at)
            or (samples_prepared_at and samples_prepared_at < record.ordered_at)
            or (samples_sequenced_at and samples_sequenced_at < record.ordered_at)
        )

    @staticmethod
    def _all_samples_have_sequence_data(links: List[models.FamilySample]) -> bool:
        """Return True if all samples are external or sequenced inhouse."""
        return all((link.sample.sequenced_at or link.sample.is_external) for link in links)

    def all_samples_have_enough_reads(
        self, links: List[models.FamilySample], threshold: float
    ) -> bool:
        return all(
            (
                link.sample.reads
                > self.Application.query.filter_by(
                    id=self.ApplicationVersion.query.filter_by(
                        id=link.sample.application_version_id
                    )
                    .first()
                    .application_id
                )
                .first()
                .target_reads
                * threshold
            )
            for link in links
        )

    def analyses_to_upload(self, pipeline: str = "") -> List[models.Analysis]:
        """Fetch analyses that haven't been uploaded."""
        records = self.Analysis.query.filter(
            models.Analysis.completed_at != None, models.Analysis.uploaded_at == None
        )

        if pipeline:
            records = records.filter(models.Analysis.pipeline.ilike(f"%{pipeline}%"))

        return records

    def analyses_to_clean(self, pipeline: str = ""):
        """Fetch analyses that haven't been cleaned."""
        records = self.latest_analyses()
        records = records.filter(
            models.Analysis.uploaded_at.isnot(None),
            models.Analysis.cleaned_at.is_(None),
            models.Analysis.pipeline.ilike(f"%{pipeline}%"),
        )
        return records

    def observations_to_upload(self):
        """Fetch observations that haven't been uploaded."""

        families_q = self.Family.query.join(
            models.Analysis, models.Family.links, models.FamilySample.sample
        ).filter(models.Sample.loqusdb_id.is_(None))

        return families_q

    def observations_uploaded(self):
        """Fetch observations that have been uploaded."""

        families_q = self.Family.query.join(models.Family.links, models.FamilySample.sample).filter(
            models.Sample.loqusdb_id.isnot(None)
        )

        return families_q

    def analyses_to_deliver(self, pipeline: str = ""):
        """Fetch analyses that have been uploaded but not delivered."""
        records = (
            self.Analysis.query.join(models.Family, models.Family.links, models.FamilySample.sample)
            .filter(
                models.Analysis.uploaded_at.isnot(None),
                models.Sample.delivered_at.is_(None),
                models.Analysis.pipeline.ilike(f"%{pipeline}%"),
            )
            .order_by(models.Analysis.uploaded_at.desc())
        )

        return records

    def analyses_to_delivery_report(self, pipeline: str = "") -> Query:
        """Fetch analyses that needs the delivery report to be regenerated."""

        analyses_query = self.latest_analyses()

        analyses_query = (
            analyses_query.filter(models.Analysis.uploaded_at)
            .filter(HASTA_IN_PRODUCTION < models.Analysis.started_at)
            .join(models.Family, models.Family.links, models.FamilySample.sample)
            .filter(
                or_(
                    models.Sample.data_analysis.is_(None),
                    models.Sample.data_analysis.ilike(f"%{pipeline}%"),
                )
            )
            .filter(
                models.Sample.delivered_at.isnot(None),
                or_(
                    models.Analysis.delivery_report_created_at.is_(None),
                    and_(
                        models.Analysis.delivery_report_created_at.isnot(None),
                        models.Analysis.delivery_report_created_at < models.Sample.delivered_at,
                    ),
                ),
            )
            .order_by(models.Analysis.uploaded_at.desc())
        )
        return analyses_query

    def samples_to_deliver(self):
        """Fetch samples that have been sequenced but not delivered."""
        records = self.Sample.query.filter(
            models.Sample.sequenced_at != None,
            models.Sample.delivered_at == None,
            models.Sample.downsampled_to == None,
        )
        return records

    def samples_not_delivered(self):
        """Fetch samples not delivered."""
        records = self.Sample.query.filter(
            models.Sample.delivered_at == None, models.Sample.downsampled_to == None
        )
        return records

    def samples_not_invoiced(self):
        """Fetch all samples that are not invoiced."""
        records = self.Sample.query.filter(
            models.Sample.downsampled_to == None, models.Sample.invoice_id == None
        )
        return records

    def samples_not_downsampled(self):
        """Fetch all samples that are not down sampled."""
        records = self.Sample.query.filter(models.Sample.downsampled_to == None)
        return records

    def microbial_samples_to_invoice(self, customer: models.Customer = None):
        """Fetch microbial samples that should be invoiced.

        Returns microbial samples that have been delivered but not invoiced.
        """
        records = self.Sample.query.filter(
            "microbial" in self.Sample.data_analysis,
            models.Sample.delivered_at is not None,
            models.Sample.invoice_id == None,
        )
        customers_to_invoice = list(set([record.customer for record in records.all()]))
        if customer:
            records = records.join(models.Family).filter(models.Family.customer_id == customer.id)
        return records, customers_to_invoice

    def samples_to_invoice(self, customer: models.Customer = None):
        """Fetch samples that should be invoiced.

        Returns samples have been delivered but not invoiced, excluding those that
        have been marked to skip invoicing.
        """
        records = self.Sample.query.filter(
            models.Sample.delivered_at != None,
            models.Sample.invoice_id == None,
            models.Sample.no_invoice == False,
            models.Sample.downsampled_to == None,
        )
        customers_to_invoice = [
            record.customer
            for record in records.all()
            if not record.customer.internal_id == "cust000"
        ]
        customers_to_invoice = list(set(customers_to_invoice))
        records = records.filter(models.Sample.customer == customer) if customer else records
        return records, customers_to_invoice

    def pools_to_invoice(self, customer: models.Customer = None):
        """
        Fetch pools that should be invoiced.
        """
        records = self.Pool.query.filter(
            models.Pool.invoice_id == None,
            models.Pool.no_invoice == False,
            models.Pool.delivered_at != None,
        )

        customers_to_invoice = [
            record.customer
            for record in records.all()
            if not record.customer.internal_id == "cust000"
        ]
        customers_to_invoice = list(set(customers_to_invoice))
        records = records.filter(models.Pool.customer_id == customer.id) if customer else records
        return records, customers_to_invoice

    def pools_to_receive(self):
        """Fetch pools that have been not yet been received."""
        records = self.Pool.query.filter(models.Pool.received_at == None)
        return records

    def pools_to_deliver(self):
        """Fetch pools that have been not yet been delivered."""
        records = self.Pool.query.filter(
            models.Pool.received_at != None, models.Pool.delivered_at == None
        )
        return records

    def microbial_samples_to_receive(self, external=False):
        """Fetch microbial samples from statusdb that have no received_at date."""
        records = (
            self.MicrobialSample.query.join(
                models.MicrobialSample.application_version, models.ApplicationVersion.application
            )
            .filter(
                models.MicrobialSample.received_at == None,
                models.Application.is_external == external,
            )
            .order_by(models.MicrobialSample.created_at)
        )
        return records

    def microbial_samples_to_prepare(self, external=False):
        """Fetch microbial samples from statusdb that have no prepared_at date."""
        records = (
            self.Sample.query.join(
                models.Sample.application_version, models.ApplicationVersion.application
            )
            .filter(
                "microbial" in self.Sample.data_analysis,
                models.Sample.prepared_at == None,
                models.Application.is_external == external,
            )
            .order_by(models.Sample.created_at)
        )
        return records

    def microbial_samples_to_sequence(self, external=False):
        """Fetch microbial samples from statusdb that have no sequenced_at date."""
        records = (
            self.Sample.query.join(
                models.Sample.application_version, models.ApplicationVersion.application
            )
            .filter(
                "microbial" in self.Sample.data_analysis,
                models.Sample.sequenced_at == None,
                models.Application.is_external == external,
            )
            .order_by(models.Sample.created_at)
        )
        return records

    def microbial_samples_to_deliver(self, external=False):
        """Fetch microbial samples from statusdb that have no delivered_at date."""
        records = (
            self.Sample.query.join(
                models.Sample.application_version, models.ApplicationVersion.application
            )
            .filter(
                "microbial" in self.Sample.data_analysis,
                models.Sample.delivered_at == None,
                models.Application.is_external == external,
            )
            .order_by(models.Sample.created_at)
        )
        return records

    def _calculate_estimated_turnaround_time(
        self,
        is_rerun,
        external_case_bool,
        analysis_ordered_at,
        samples_received_at,
        samples_prepared_at,
        samples_sequenced_at,
        analysis_completed_at,
        analysis_uploaded_at,
        samples_delivered_at,
    ):
        """Calculated estimated turnaround-time"""
        if samples_received_at and samples_delivered_at:
            return self._calculate_date_delta(None, samples_received_at, samples_delivered_at)

        o_a = self._calculate_date_delta(5, analysis_ordered_at, analysis_completed_at)
        r_p = self._calculate_date_delta(4, samples_received_at, samples_prepared_at)
        p_s = self._calculate_date_delta(5, samples_prepared_at, samples_sequenced_at)
        s_a = self._calculate_date_delta(4, samples_sequenced_at, analysis_completed_at)
        a_u = self._calculate_date_delta(1, analysis_completed_at, analysis_uploaded_at)
        u_d = self._calculate_date_delta(2, analysis_uploaded_at, samples_delivered_at)

        if is_rerun:
            o_a = self._calculate_date_delta(1, analysis_ordered_at, analysis_completed_at)
            return o_a + a_u

        if external_case_bool:
            if analysis_ordered_at and analysis_uploaded_at:
                return self._calculate_date_delta(None, analysis_ordered_at, analysis_uploaded_at)

            return o_a + a_u

        if samples_received_at and samples_delivered_at:
            return self._calculate_date_delta(None, samples_received_at, samples_delivered_at)

        return r_p + p_s + s_a + a_u + u_d

    @staticmethod
    def _calculate_date_delta(default, first_date, last_date):
        # calculates date delta between two dates, assumes last_date is today if missing
        delta = default
        if not last_date:
            last_date = datetime.now()
        if first_date:
            delta = (last_date - first_date).days
        return delta

    @staticmethod
    def _get_max_tat(links):
        max_tat = 0
        for link in links:
            if link.sample.application_version.application.turnaround_time:
                max_tat = max(0, link.sample.application_version.application.turnaround_time)
        return max_tat



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import GenericRepr, Snapshot


snapshots = Snapshot()

snapshots["test_parse_sampleinfo_rna_result_contents 1"] = {
    "bcftools_merge": "/path/to/stuff/rare-disease/cases/case_id/analysis/files/case_id_comb.vcf",
    "case": "case_id",
    "config_file_path": "/path/to/stuff/rare-disease/cases/case_id/analysis/case_id_config.yaml",
    "date": GenericRepr("datetime.datetime(2019, 11, 21, 14, 31, 2)"),
    "is_finished": True,
    "multiqc_html": "/path/to/stuff/rare-disease/cases/case_id/analysis/files/multiqc_report.html",
    "multiqc_json": "/path/to/stuff/rare-disease/cases/case_id/analysis/files/multiqc_data.json",
    "pedigree_path": "/path/to/stuff/rare-disease/cases/case_id/analysis/files/case_id.fam",
    "qcmetrics_path": (
        "/path/to/stuff/rare-disease/cases/case_id/analysis/files/qcmetrics_file.stub"
    ),
    "samples": [
        {
            "bam": "/path/to/stuff/rare-disease/cases/case_id/analysis/files/sample_id_1_lanes_1234_trim_star_sorted_merged_md_splitncigar_brecal.bam",
            "bootstrap_vcf": [
                "/path/to/stuff/rare-disease/cases/case_id/analysis/files/foo_sample_id_1_lanes_1234_trim_star_sorted_merged_md_splitncigar_brecal_haptc_filtered_ase_bootstr.vcf",
                "/path/to/stuff/rare-disease/cases/case_id/analysis/files/bar_sample_id_1_lanes_1234_trim_star_sorted_merged_md_splitncigar_brecal_haptc_filtered_ase_bootstr.vcf",
            ],
            "gatk_asereadcounter": [
                "/path/to/stuff/rare-disease/cases/case_id/analysis/files/sample_id_1_lanes_1234_trim_star_sorted_merged_md_splitncigar_brecal_haptc_filtered_ase.csv"
            ],
            "gatk_baserecalibration": [
                "/path/to/stuff/rare-disease/cases/case_id/analysis/files/sample_id_1_lanes_1234_trim_star_sorted_merged_md_splitncigar_brecal.bam"
            ],
            "gffcompare_ar": [
                "/path/to/stuff/rare-disease/cases/case_id/analysis/files/sample_id_1_lanes_1234_trim_star_sorted_merged_strg_gffcmp.gtf"
            ],
            "id": "sample_id_1",
            "mark_duplicates": [
                "/path/to/stuff/rare-disease/cases/case_id/analysis/files/sample_id_1_lanes_1234_trim_star_sorted_merged_md_metric"
            ],
            "salmon_quant": ["/path/to/stuff/rare-disease/cases/case_id/analysis/files/quant.sf"],
            "star_fusion": "/path/to/stuff/rare-disease/cases/case_id/analysis/files/star-fusion.fusion_predictions.tsv",
            "stringtie_ar": [
                "/path/to/stuff/rare-disease/cases/case_id/analysis/files/sample_id_1_lanes_1234_trim_star_sorted_merged_strg.gtf"
            ],
        }
    ],
    "vep_path": "/path/to/stuff/rare-disease/cases/case_id/analysis/files/case_id_comb_vep.vcf",
    "version": "v7.1.4",
    "version_collect_ar_path": (
        "/path/to/stuff/rare-disease/cases/case_id/analysis/files/case_id_vcol.yaml"
    ),
}



"""Module to create MIP analysis delivery reports"""
import logging
from datetime import datetime
from pathlib import Path

import requests
import ruamel.yaml
from jinja2 import Environment, PackageLoader, select_autoescape
from cg.apps.coverage import ChanjoAPI
from cg.apps.lims import LimsAPI
from cg.apps.scoutapi import ScoutAPI
from cg.meta.report.report_validator import ReportValidator
from cg.meta.workflow.mip import AnalysisAPI
from cg.meta.report.sample_calculator import SampleCalculator
from cg.meta.report.report_helper import ReportHelper
from cg.meta.report.presenter import Presenter
from cg.store import Store, models
from cg.exc import DeliveryReportError


class ReportAPI:
    """API to create MIP analysis delivery reports"""

    def __init__(
        self,
        store: Store,
        lims_api: LimsAPI,
        chanjo_api: ChanjoAPI,
        analysis_api: AnalysisAPI,
        scout_api: ScoutAPI,
        logger=logging.getLogger(__name__),
        yaml_loader=ruamel.yaml,
        path_tool=Path,
    ):

        self.store = store
        self.lims = lims_api
        self.chanjo = chanjo_api
        self.analysis = analysis_api
        self.log = logger
        self.yaml_loader = yaml_loader
        self.path_tool = path_tool
        self.scout = scout_api
        self.report_validator = ReportValidator(store)

    def create_delivery_report(self, case_id: str, accept_missing_data: bool = False) -> str:
        """Generate the html contents of a delivery report."""

        delivery_data = self._get_delivery_data(case_id)
        self._handle_missing_report_data(accept_missing_data, delivery_data, case_id)
        report_data = self._make_data_presentable(delivery_data)
        self._handle_missing_report_data(accept_missing_data, report_data, case_id)
        rendered_report = self._render_delivery_report(report_data)
        return rendered_report

    def _handle_missing_report_data(self, accept_missing_data, delivery_data, case_id):
        """Handle when some crucial data is missing in the report"""
        if not self.report_validator.has_required_data(delivery_data):
            if not accept_missing_data:
                raise DeliveryReportError(
                    f"Could not generate report data for {case_id}, "
                    f"missing data:"
                    f" {self.report_validator.get_missing_attributes()}"
                )

            self.log.warning(
                "missing data: %s", ", ".join(self.report_validator.get_missing_attributes())
            )

    def create_delivery_report_file(
        self, case_id: str, file_path: Path, accept_missing_data: bool = False
    ):
        """Generate a temporary file containing a delivery report."""

        delivery_report = self.create_delivery_report(
            case_id=case_id, accept_missing_data=accept_missing_data
        )

        file_path.mkdir(parents=True, exist_ok=True)
        report_file_path = Path(file_path / "delivery-report.html")
        delivery_report_file = open(report_file_path, "w")
        delivery_report_file.write(delivery_report)
        delivery_report_file.close()

        return delivery_report_file

    def _get_delivery_data(self, case_id: str) -> dict:
        """Fetch all data needed to render a delivery report."""

        report_data = dict()
        case_obj = self._get_case_from_statusdb(case_id)
        analysis_obj = case_obj.analyses[0]

        report_data["case"] = case_obj.name
        report_data["pipeline"] = analysis_obj.pipeline
        report_data["pipeline_version"] = analysis_obj.pipeline_version
        report_data["customer_name"] = case_obj.customer.name
        report_data["customer_internal_id"] = case_obj.customer.internal_id
        report_data["customer_invoice_address"] = case_obj.customer.invoice_address
        report_data["scout_access"] = bool(case_obj.customer.scout_access)
        report_data["report_version"] = ReportHelper.get_report_version(analysis_obj)
        report_data["previous_report_version"] = ReportHelper.get_previous_report_version(
            analysis_obj
        )

        report_data["samples"] = self._fetch_case_samples_from_status_db(case_id)
        report_data["panels"] = self._fetch_panels_from_status_db(case_id)
        self._incorporate_lims_data(report_data)
        self._incorporate_lims_methods(report_data["samples"])
        self._incorporate_coverage_data(report_data["samples"], report_data["panels"])
        self._incorporate_trending_data(report_data, case_id)

        report_data["today"] = datetime.today()
        application_data = self._get_application_data_from_status_db(report_data["samples"])
        report_data.update(application_data)
        return report_data

    def _get_case_from_statusdb(self, case_id: str) -> models.Family:
        """Fetch a case object from the status database."""
        return self.store.family(case_id)

    def _incorporate_lims_methods(self, samples: list):
        """Fetch the methods used for preparation, sequencing and delivery of the samples."""

        for sample in samples:
            lims_id = sample["internal_id"]
            method_types = ["prep_method", "sequencing_method"]
            for method_type in method_types:
                get_method = getattr(self.lims, f"get_{method_type}")
                method_name = get_method(lims_id)
                sample[method_type] = method_name

    @staticmethod
    def _render_delivery_report(report_data: dict) -> str:
        """Render and return report data on the report Jinja template."""

        env = Environment(
            loader=PackageLoader("cg", "meta/report/templates"),
            autoescape=select_autoescape(["html", "xml"]),
        )
        template = env.get_template("report.html")
        template_out = template.render(**report_data)
        return template_out

    def _get_sample_coverage_from_chanjo(self, lims_id: str, genes: list) -> dict:

        """Get coverage data from Chanjo for a sample."""
        return self.chanjo.sample_coverage(lims_id, genes)

    def _incorporate_trending_data(self, report_data: dict, case_id: str):
        """Incorporate trending data into a set of samples."""
        trending_data = self.analysis.get_latest_metadata(family_id=case_id)

        mapped_reads_all_samples = trending_data.get("mapped_reads", {})
        duplicates_all_samples = trending_data.get("duplicates", {})
        analysis_sex_all_samples = trending_data.get("analysis_sex", {})

        for sample in report_data["samples"]:
            lims_id = sample["internal_id"]
            sample["analysis_sex"] = analysis_sex_all_samples.get(lims_id)
            sample["mapped_reads"] = mapped_reads_all_samples.get(lims_id)
            sample["duplicates"] = duplicates_all_samples.get(lims_id)

        report_data["genome_build"] = trending_data.get("genome_build")

    def _incorporate_coverage_data(self, samples: list, panels: list):
        """Incorporate coverage data from Chanjo for each sample ."""

        genes = self._get_genes_from_scout(panels)

        for sample in samples:
            lims_id = sample["internal_id"]

            sample_coverage = self._get_sample_coverage_from_chanjo(lims_id, genes)

            target_coverage = None
            target_completeness = None

            if sample_coverage:
                target_coverage = sample_coverage.get("mean_coverage")
                target_completeness = sample_coverage.get("mean_completeness")
            else:
                self.log.warning("No coverage could be calculated for: %s", lims_id)

            sample["target_coverage"] = target_coverage
            sample["target_completeness"] = target_completeness

    def _fetch_case_samples_from_status_db(self, case_id: str) -> list:
        """Incorporate data from the status database for each sample ."""

        delivery_data_samples = list()
        case_samples = self.store.family_samples(case_id)

        for case_sample in case_samples:
            sample = case_sample.sample
            delivery_data_sample = dict()
            delivery_data_sample["internal_id"] = sample.internal_id
            delivery_data_sample["ticket"] = sample.ticket_number
            delivery_data_sample["status"] = case_sample.status
            delivery_data_sample["received_at"] = sample.received_at
            delivery_data_sample["prepared_at"] = sample.prepared_at
            delivery_data_sample["sequenced_at"] = sample.sequenced_at
            delivery_data_sample["delivered_at"] = sample.delivered_at
            delivery_data_sample["processing_time"] = SampleCalculator.calculate_processing_days(
                sample
            )
            delivery_data_sample["ordered_at"] = sample.ordered_at
            delivery_data_sample["million_read_pairs"] = (
                round(sample.reads / 2000000, 1) if sample.reads else None
            )

            delivery_data_sample["capture_kit"] = sample.capture_kit
            delivery_data_sample["data_analysis"] = sample.data_analysis
            delivery_data_samples.append(delivery_data_sample)

        return delivery_data_samples

    def _get_application_data_from_status_db(self, samples: list) -> dict:
        """Fetch application data including accreditation status for all samples."""
        application_data = dict()
        used_applications = set()

        for sample in samples:
            used_applications.add((sample["application"], sample["application_version"]))

        applications = []
        accreditations = []
        for apptag_id, apptag_version in used_applications:

            application_is_accredited = False

            application = self.store.application(tag=apptag_id)

            if application:
                application_is_accredited = application.is_accredited
                applications.append(
                    {
                        "tag": application.tag,
                        "description": application.description,
                        "limitations": application.limitations,
                    }
                )

            accreditations.append(application_is_accredited)

        application_data["applications"] = applications
        application_data["accredited"] = bool(all(accreditations))
        return application_data

    def _fetch_panels_from_status_db(self, case_id: str) -> list:
        """fetch data from the status database for each panels ."""
        case_obj = self.store.family(case_id)
        panels = case_obj.panels
        return panels

    def _incorporate_lims_data(self, report_data: dict):
        """Incorporate data from LIMS for each sample ."""
        for sample in report_data.get("samples"):
            lims_id = sample["internal_id"]

            try:
                lims_sample = self.lims.sample(lims_id)
            except requests.exceptions.HTTPError as error:
                lims_sample = dict()
                self.log.info("could not fetch sample %s from LIMS: %s", lims_id, error)

            sample["name"] = lims_sample.get("name")
            sample["sex"] = lims_sample.get("sex")
            sample["source"] = lims_sample.get("source")
            sample["application"] = lims_sample.get("application")
            sample["application_version"] = lims_sample.get("application_version")

    def _get_genes_from_scout(self, panels: list) -> list:
        panel_genes = list()

        for panel in panels:
            panel_genes.extend(self.scout.get_genes(panel))

        panel_gene_ids = [gene.get("hgnc_id") for gene in panel_genes]

        return panel_gene_ids

    @staticmethod
    def _make_data_presentable(delivery_data: dict) -> dict:
        """Replace db values with what a human might expect"""

        presenter = Presenter(precision=2)

        _presentable_dict = presenter.process_dict(delivery_data)

        for sample in delivery_data["samples"]:
            sample["mapped_reads"] = presenter.process_float_string(sample["mapped_reads"], 2)
            sample["target_coverage"] = presenter.process_float_string(sample["target_coverage"], 1)
            sample["target_completeness"] = presenter.process_float_string(
                sample["target_completeness"], 2
            )
            sample["duplicates"] = presenter.process_float_string(sample["duplicates"], 1)

        return _presentable_dict



"""Tests for reset part of the store API"""

from datetime import datetime, timedelta

from cg.store import Store


def test_reset_observation(store: Store):
    # GIVEN a store with a case with loqus-links
    family = add_family(store)
    sample = add_sample(store, loqusdb_links=True)
    store.relate_sample(family=family, sample=sample, status="unknown")
    assert sample.loqusdb_id is not None

    # WHEN calling reset observations
    store.reset_observations(case_id=family.internal_id)

    # THEN the links to observations in loqusdb should have been reset
    assert sample.loqusdb_id is None


def ensure_application_version(disk_store, application_tag="dummy_tag"):
    """utility function to return existing or create application version for tests"""
    application = disk_store.application(tag=application_tag)
    if not application:
        application = disk_store.add_application(
            tag=application_tag,
            category="wgs",
            description="dummy_description",
            percent_kth=80,
        )
        disk_store.add_commit(application)

    prices = {"standard": 10, "priority": 20, "express": 30, "research": 5}
    version = disk_store.application_version(application, 1)
    if not version:
        version = disk_store.add_version(application, 1, valid_from=datetime.now(), prices=prices)

        disk_store.add_commit(version)
    return version


def ensure_customer(disk_store, customer_id="cust_test"):
    """utility function to return existing or create customer for tests"""
    customer_group = disk_store.customer_group("dummy_group")
    if not customer_group:
        customer_group = disk_store.add_customer_group("dummy_group", "dummy group")

        customer = disk_store.add_customer(
            internal_id=customer_id,
            name="Test Customer",
            scout_access=False,
            customer_group=customer_group,
            invoice_address="dummy_address",
            invoice_reference="dummy_reference",
        )
        disk_store.add_commit(customer)
    customer = disk_store.customer(customer_id)
    return customer


def add_sample(store, sample_name="sample_test", loqusdb_links=False):
    """utility function to add a sample to use in tests"""

    customer = ensure_customer(store)
    application_version_id = ensure_application_version(store).id
    sample = store.add_sample(name=sample_name, sex="unknown")
    sample.application_version_id = application_version_id
    sample.customer = customer
    if loqusdb_links:
        sample.loqusdb_id = True
    store.add_commit(sample)
    return sample


def ensure_panel(disk_store, panel_id="panel_test", customer_id="cust_test"):
    """utility function to add a panel to use in tests"""
    customer = ensure_customer(disk_store, customer_id)
    panel = disk_store.panel(panel_id)
    if not panel:
        panel = disk_store.add_panel(
            customer=customer,
            name=panel_id,
            abbrev=panel_id,
            version=1.0,
            date=datetime.now(),
            genes=1,
        )
        disk_store.add_commit(panel)
    return panel


def add_family(
    disk_store,
    family_id="family_test",
    customer_id="cust_test",
    ordered_days_ago=0,
    action=None,
    priority=None,
):
    """utility function to add a family to use in tests"""
    panel = ensure_panel(disk_store)
    customer = ensure_customer(disk_store, customer_id)
    family = disk_store.add_family(name=family_id, panels=panel.name)
    family.customer = customer
    family.ordered_at = datetime.now() - timedelta(days=ordered_days_ago)
    if action:
        family.action = action
    if priority:
        family.priority = priority
    disk_store.add_commit(family)
    return family
